# -*- coding: utf-8 -*-

#
# Override bl_ui/properties_object.py
#

import bpy
import Spr
import time

import spb.properties
import spb.ui

from   spb.properties    import register_properties
from   spb.handlers      import Handlers
from   spb.spbapi        import spbapi
from   mathutils         import Vector,Quaternion

from   spb.utils         import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handlers
#

handlers = Handlers()


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Operator
#

class SprOperator(bpy.types.Operator):
	'''SprPyBlender Operator'''
	bl_idname = "spr.operator"
	bl_label  = "SprOperator"

	timer = None
	step_count = 0
	prev_time  = time.time()
	
	def modal(self, context, event):
		if event.type == 'TIMER':
			# BpyObjectのキャッシュをクリア
			BpyObjectRefresh()

			# ハンドラの処理を行う
			handlers.step()

			# CPSを計測する
			self.step_count += 1
			curr_time  = time.time()
			time_delta = curr_time - self.prev_time
			if time_delta > 1.0:
				self.prev_time = curr_time
				context.scene.spr_op_cps_count = (self.step_count / time_delta)
				self.step_count = 0

			context.scene.spr_phys_cps_count = spbapi.GetCPS()

		return {'PASS_THROUGH'}

	def execute(self, context):
		#timer登録
		context.window_manager.modal_handler_add(self)
		self.timer = context.window_manager.event_timer_add(1/100.0, context.window)
		return {'RUNNING_MODAL'}

	def cancel(self, context):
		context.window_manager.event_timer_remove(self.timer)
		return {'CANCELLED'}




# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# SprOperatorパネルのレイアウト
#

class SprBlenderInfoPanel(bpy.types.Panel):
	bl_label = "SprBlender Info"
	bl_space_type = "VIEW_3D"
	bl_region_type = "UI"

	def draw(self, context):
		layout = self.layout
		scene = context.scene

		layout.label("Frame Rate")
		layout.prop(scene, "spr_op_cps_count",   text="Operator", emboss=False)
		layout.prop(scene, "spr_phys_cps_count", text="Physics",  emboss=False)


simulation_started = False
class SprOperatorPanel(bpy.types.Panel):
	bl_label = "SprBlender"
	bl_space_type = "VIEW_3D"
	bl_region_type = "UI"

	def draw(self, context):
		layout = self.layout
		scene = context.scene
		
		col = layout.column(align=True)
		col.label(text="Simulation:")

		row = col.row(align=True)
		if scene.spr_step_enabled==0:
			if spb.simulation_started:
				text = "ReStart"
				row.alert = True
			else:
				row.alert = False
				text = "Start"
			icon = "PLAY"
		else:
			row.alert = True
			text = "Pause"
			icon = "PAUSE"
		row.prop(scene, "spr_step_enabled", text=text, icon=icon)
		col.operator("spr.stop_simulation", text="Stop", icon = "MESH_PLANE")

		row = col.row(align=True)
		row.operator("spr.quickrestore_to_started", text="Back", icon = "REW")
		row.operator("spr.quickrestore_to_ended",   text="End", icon = "FF")

		col.operator("spr.one_step",text = "One Step" ,icon="FRAME_NEXT")
		
		col = layout.column(align=True)
		col.prop(scene, "spr_record", text = "Record", icon="REC")
		if scene.spr_record:
			col.prop(scene, "spr_recording_frame", text = "Frame")
			col.prop(scene, "spr_record_every_n_steps", text = "Rec Per Step")

			row = col.row(align=True)
			row.prop(scene, "spr_record_to_cache", text = "Use Cache", toggle=True)
			if scene.spr_record_to_cache:
				row.operator("spr.bake_keyframes",text = "Bake")

			# col.prop(scene, "spr_step_frame", text = "Step Frame No.")
			# col.prop(scene, "spr_sync_current_frame", text = "Sync Current Frame")

		# --
		
		col = layout.column(align=True)
		col.label(text="Restore / Store:")
		col.operator("spr.restore",text = "Restore Scene", icon="RECOVER_LAST")

		col = layout.column(align=True)
		col.operator("spr.store",text = "Store Scene")
		
		# --
		
		col = layout.column(align=True)
		col.label(text="Rebuild:")
		col.operator("spr.rebuild",text = "Rebuild Scene", icon="SCENE_DATA")

		# --
		
		col = layout.column(align=True)
		col.label(text = "Enable Functions:")

		row = col.row(align=True)
		row.prop(scene, "spr_minimal_sync",      text = "Minimal Sync")
		row.operator("spr.synchronize",text = "Sync")

		col.prop(scene, "spr_throw_enabled",     text = "Throw")
		col.prop(scene, "spr_ik_engine_enabled", text = "IK Engine")

		row = col.row(align=True)
		row.prop(scene, "spr_run_scripts_enabled", text = "Script")
		row.operator("spr.update_scripts",text = "Update")

		row = col.row(align=True)
		row.prop(scene, "spr_creature_enabled",     text = "Creature")
		row.operator("spr.update_creature_scripts", text = "Update")

		# show_icons(layout)


class SprBlenderAdvancedPanel(bpy.types.Panel):
	bl_label = "SprBlender Advanced"
	bl_space_type = "VIEW_3D"
	bl_region_type = "UI"

	def draw(self, context):
		layout = self.layout
		scene = context.scene
		col = layout.column()
		col.prop(scene, "spr_use_object_wrapper", text = "Assure Availability")
		col.operator("spr.calibrate_spidar",      text = "Calibrate SPIDAR")
		if scene.spr_need_migration==1:
			col = layout.column(align=True)
			col.operator("spr.migration",text = "Migration 2.0")


# --- --- --- --- --- --- --- --- --- --- 
# ボタンを押した際の動作

class OBJECT_OT_SprOneStep(bpy.types.Operator):
	bl_idname = "spr.one_step"
	bl_label = "[spr]Call One Step"
	bl_description = "[Spr]Call ONE physics step."
	
	def execute(self,context):
		spb.handlers.store("quickstore_on_start", init=True)
		spb.spbapi.GetFWSdk().GetScene(0).GetPHScene().Step()
		spb.spbapi.GetCRSdk().Step()
		spb.handlers.step()
		spb.handlers.store("quickstore_on_end")
		
		return{'FINISHED'}

class OBJECT_OT_SprStopSimulation(bpy.types.Operator):
	bl_idname = "spr.stop_simulation"
	bl_label = "[spr]Stop Simulation"
	bl_description = "[Spr]Stop Simulation."
	
	def execute(self,context):
		spb.simulation_started = False
		context.scene.spr_step_enabled = 0
		
		return{'FINISHED'}

class OBJECT_OT_SprBakeKeyframes(bpy.types.Operator):
	bl_idname = "spr.bake_keyframes"
	bl_label = "[spr]Bake Keyframes"
	bl_description = "[Spr]Bake Cached Keyframes into Blender Keys."
	
	def execute(self,context):
		spb.handlers.bake_frame = 0
		spb.handlers.bake_mode  = True
		
		return{'FINISHED'}

class OBJECT_OT_SprStore(bpy.types.Operator):
	bl_idname = "spr.store"
	bl_label = "[spr]Store scene"
	bl_description = "[Spr]Store pose of objects in the scene."
	
	def execute(self,context):
		spb.handlers.store("default")
		
		return{'FINISHED'}

class OBJECT_OT_SprRestore(bpy.types.Operator):
	bl_idname = "spr.restore"
	bl_label = "[spr]Restore scene"
	bl_description = "[Spr]Restore pose of objects in the scene to stored state."
	
	def execute(self,context):
		spb.handlers.restore("default")
		
		return{'FINISHED'}
		
class OBJECT_OT_SprQuickRestoreToStarted(bpy.types.Operator):
	bl_idname = "spr.quickrestore_to_started"
	bl_label = "[spr] Quick Restore (started)"
	bl_description = "[Spr]Restore Scene to the state before simulation started."
	
	def execute(self,context):
		spb.handlers.restore("quickstore_on_start")
		
		return{'FINISHED'}

class OBJECT_OT_SprQuickRestoreToEnded(bpy.types.Operator):
	bl_idname = "spr.quickrestore_to_ended"
	bl_label = "[spr] Quick Restore (stopped)"
	bl_description = "[Spr]Restore Scene to the state after simulation stopped."
	
	def execute(self,context):
		spb.handlers.restore("quickstore_on_end")
		
		return{'FINISHED'}

class OBJECT_OT_SprRebuild(bpy.types.Operator):
	bl_idname = "spr.rebuild"
	bl_label = "[spr]Rebuild scene"
	bl_description = "Clear Springhead Scene and Rebuild from bpy scene."
	
	def execute(self,context):
		spb.handlers.operation = False

		# ハンドラの初期化
		spb.handlers.clear()
		# Springheadの初期化
		spb.spbapi.Init()

		spb.handlers.operation = True
		
		return{'FINISHED'}

class OBJECT_OT_SPBSynchronize(bpy.types.Operator):
	bl_idname = "spr.synchronize"
	bl_label = "[spr]Synchronize Step"
	bl_description = "[Spr]Synchronize Step"
	
	def execute(self,context):
		spb.handlers.full_step()
		
		return{'FINISHED'}

class OBJECT_OT_UpdateCreatureScripts(bpy.types.Operator):
	bl_idname = "spr.update_creature_scripts"
	bl_label = "[spr]Update Creature Scripts"
	bl_description = "[Spr]Update creature scripts."
	
	def execute(self,context):
		for creature_handler in spb.handlers.creature_handlers.values():
			creature_handler.rule.update()
		
		return{'FINISHED'}

class OBJECT_OT_UpdateScripts(bpy.types.Operator):
	bl_idname = "spr.update_scripts"
	bl_label = "[spr]Update Scripts"
	bl_description = "[Spr]Update user scripts."
	
	def execute(self,context):
		spb.handlers.scene_handler.rule.update()
		
		return{'FINISHED'}

class OBJECT_OT_SprCalibrateSpidar(bpy.types.Operator):
	bl_idname = "spr.calibrate_spidar"
	bl_label = "[spr]Calibrate SPIDAR"
	bl_description = "[Spr]Calibrate SPIDAR."
	
	def execute(self,context):
		spbapi.GetSPIDAR(0).Calibration()
		spbapi.GetSPIDAR(1).Calibration()
		
		return{'FINISHED'}

class OBJECT_OT_SprMigration(bpy.types.Operator):
	bl_idname = "spr.migration"
	bl_label = "[spr]Migration to SprBlender 2.0"
	bl_description = "Migrate Properties Definition into SprBlender 2.0 style."
	
	def execute(self,context):
		for obj in bpy.data.objects:
			# Make Joint
			# Jointなのにchildに p_ が存在しない場合は，追加してrotation_quaternion, locationをコピっておく
			if (not obj.spr_joint_type=="None" and obj.spr_additional_joint==0):
				found=False
				for child in obj.children:
					if child.name==("p_"+obj.name):
						found=True
				if not found:
					obj_jo = bpy.data.objects.new("p_"+obj.name, None)
					context.scene.objects.link(obj_jo)
					obj_jo.empty_draw_type = "ARROWS"
					obj_jo.parent = obj
					obj_jo.show_x_ray = True
					bpy.ops.object.select_all(action='DESELECT')
					obj_jo.select = True
					bpy.ops.object.parent_clear(type="CLEAR_INVERSE")
					obj_jo.rotation_mode="QUATERNION"
					##plugオブジェクトが無くても機能してるjointは存在するので
					#obj_jo.rotation_quaternion = Quaternion((1,0,0,0))
					#obj_jo.location = Vector((0,0,0))
					obj_jo.location = Vector((obj.spr_joint_plug_pos_x, obj.spr_joint_plug_pos_y, obj.spr_joint_plug_pos_z))
					obj_jo.rotation_quaternion = Quaternion((obj.spr_joint_plug_ori_w, obj.spr_joint_plug_ori_x, obj.spr_joint_plug_ori_y, obj.spr_joint_plug_ori_z))

		for obj in bpy.data.objects:
			# Solid
			if obj.spr_physics_type=="Dynamic" or (not obj.spr_shape_type=="None"):
				obj.spr_object_type     = "Solid"
				obj.spr_solid_static    = 1 if (obj.spr_physics_type=="Static") else 0
				obj.spr_autoset_inertia = (not obj.spr_inertia_manual_setting_enabled)

			# Joint
			if (not obj.parent is None) and (("p_" + obj.parent.name) == obj.name) or (obj.spr_additional_joint==1):
				obj.spr_object_type         = "Joint"
				obj.spr_joint_socket_target = "Solid Name"

				if obj.spr_additional_joint==0:
					object = obj.parent
				else:
					object = obj

				# オブジェクトが違う（かもしれない）のでJointのプロパティは全部コピる必要がある
				obj.spr_joint_mode      = object.spr_joint_type
				obj.spr_joint_enabled   = object.spr_joint_enabled
				obj.spr_spring          = object.spr_spring
				obj.spr_damper          = object.spr_damper
				obj.spr_joint_max_force = object.spr_joint_max_force
				obj.spr_joint_collision = object.spr_joint_collision

				obj.spr_joint_socket_target_name = object.spr_joint_target
				obj.hide_select = False

				if object.spr_joint_type=="Ball":
					if (object.spr_limit_swing_min < 3e+38) or (object.spr_limit_swing_max < 3e+38):
						obj.spr_joint_limit_enable = True
						obj.spr_limit_swing_min = object.spr_limit_swing_min
						obj.spr_limit_swing_max = object.spr_limit_swing_max
					if (object.spr_limit_twist_min < 3e+38) or (object.spr_limit_twist_max < 3e+38):
						obj.spr_joint_limit_enable = True
						obj.spr_limit_twist_min = object.spr_limit_twist_min
						obj.spr_limit_twist_max = object.spr_limit_twist_max

				if object.spr_joint_type=="Hinge" or object.spr_joint_type=="Slider":
					if (object.spr_limit_range_min < 3e+38) or (object.spr_limit_range_max < 3e+38):
						obj.spr_joint_limit_enable = True
						obj.spr_limit_range_min = object.spr_limit_range_min
						obj.spr_limit_range_max = object.spr_limit_range_max

				obj.spr_limit_spring = object.spr_limit_spring
				obj.spr_limit_damper = object.spr_limit_damper

				obj.spr_joint_target_position   = object.spr_joint_target_position
				obj.spr_joint_target_position_x = object.spr_joint_target_position_x
				obj.spr_joint_target_position_y = object.spr_joint_target_position_y
				obj.spr_joint_target_position_z = object.spr_joint_target_position_z

				obj.spr_joint_target_velocity   = object.spr_joint_target_velocity
				obj.spr_joint_target_velocity_x = object.spr_joint_target_velocity_x
				obj.spr_joint_target_velocity_y = object.spr_joint_target_velocity_y
				obj.spr_joint_target_velocity_z = object.spr_joint_target_velocity_z

				obj.spr_joint_offset_force      = object.spr_joint_offset_force
				obj.spr_joint_offset_force_x    = object.spr_joint_offset_force_x
				obj.spr_joint_offset_force_y    = object.spr_joint_offset_force_y
				obj.spr_joint_offset_force_z    = object.spr_joint_offset_force_z

				obj.spr_ik_enabled              = object.spr_ik_enabled
				obj.spr_ik_bias                 = object.spr_ik_bias

		for obj in bpy.data.objects:
			# Solid
			if obj.spr_physics_type=="Dynamic" or (not obj.spr_shape_type=="None"):
				obj.spr_ik_enabled      = obj.spr_ik_end_effector

		return{'FINISHED'}


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Start Operation
#

def startup():
	'''ファイルロード時，オペレータを再登録'''
	# 登録されたものがあれば一度登録解除
	bpy.utils.unregister_module(__name__)

	# オペレータ及びプロパティの登録
	bpy.utils.register_module(__name__)
	register_properties()

	# ハンドラの初期化
	spb.handlers.clear()

	# Springheadの初期化
	spb.spbapi.Init()

	# オペレータ有効化
	bpy.ops.spr.operator()

	# 次回同じファイルをロードしたときには自動でimport spb & startupする
	if not "start_spb.py" in bpy.data.texts:
		textedit = bpy.data.texts.new("start_spb.py")
		s = ""
		s += "import bpy\n"
		s += "def on_load(context):\n"
		s += "\timport spb\n"
		s += "\tspb.startup()\n"
		s += "bpy.app.handlers.load_post.append(on_load)\n"
		textedit.from_string(s)
		textedit.use_module = True

# はじめてimport spbしたファイルだった場合，ここでstartupする
# （start_spb.pyがRegisterでない場合も）
if (not "start_spb.py" in bpy.data.texts) or (not bpy.data.texts["start_spb.py"].use_module):
	startup()
