# -*- coding: utf-8 -*-

import bpy
import spb
import time

from   copy                         import deepcopy

from   spb.handlers_imp.scene       import SceneHandler
from   spb.handlers_imp.solid       import SolidHandler
from   spb.handlers_imp.pointer     import PointerHandler
from   spb.handlers_imp.shape       import ShapeHandler
from   spb.handlers_imp.joint       import JointHandler
from   spb.handlers_imp.limit       import LimitHandler
from   spb.handlers_imp.ikeff       import IKEffHandler
from   spb.handlers_imp.ikact       import IKActHandler
from   spb.handlers_imp.creature    import CreatureHandler
from   spb.handlers_imp.crbone      import CRBoneHandler
from   spb.handlers_imp.visualizer  import VisualizerHandler

from   spb.utils                    import BpyObject, BpyGroup, NoneCheckDict, profiler, to_bpy, copy
from   spb.spbapi                   import spbapi
from   spb.keyframebin              import KeyframeBin

from   Spr                          import *


class Handlers():
	'''ハンドラの一覧を管理するクラス．'''

	def __init__(self):
		self.is_initialized      = False
		self.scene_handler       = SceneHandler(bpy.context.scene)
		self.solid_handlers      = NoneCheckDict()
		self.shape_handlers      = NoneCheckDict()
		self.joint_handlers      = NoneCheckDict()
		self.limit_handlers      = NoneCheckDict()
		self.ikeff_handlers      = NoneCheckDict()
		self.ikact_handlers      = NoneCheckDict()
		self.creature_handlers   = NoneCheckDict()
		self.crbone_handlers     = NoneCheckDict()
		self.visualizer_handlers = NoneCheckDict()

		self.handlers_list = []
		self.handlers_list.append( self.solid_handlers    )
		self.handlers_list.append( self.shape_handlers    )
		self.handlers_list.append( self.joint_handlers    )
		self.handlers_list.append( self.limit_handlers    )
		self.handlers_list.append( self.ikeff_handlers    )
		self.handlers_list.append( self.ikact_handlers    )
		self.handlers_list.append( self.creature_handlers )
		self.handlers_list.append( self.crbone_handlers )
		self.handlers_list.append( self.visualizer_handlers )

		self.last_step_enabled = False

		# これをFalseにするとHandlerの実行を一時的に止めることができる．
		self.operation = True

		# Keyframe Recorder
		self.bake_mode  = False
		self.bake_frame_no = 0
		self.keys = KeyframeBin()
		self.record_step_count = 0


	# --- --- --- --- ---

	def add_handler(self, handler):
		if type(handler) is PointerHandler:
			self.solid_handlers[handler.name] = handler
		if type(handler) is SolidHandler:
			self.solid_handlers[handler.name] = handler
		if type(handler) is ShapeHandler:
			self.shape_handlers[handler.name] = handler
		if type(handler) is JointHandler:
			self.joint_handlers[handler.name] = handler
		if type(handler) is LimitHandler:
			self.limit_handlers[handler.name] = handler
		if type(handler) is IKEffHandler:
			self.ikeff_handlers[handler.name] = handler
		if type(handler) is IKActHandler:
			self.ikact_handlers[handler.name] = handler
		if type(handler) is CreatureHandler:
			self.creature_handlers[handler.name] = handler
		if type(handler) is CRBoneHandler:
			self.crbone_handlers[handler.name] = handler
		if type(handler) is VisualizerHandler:
			self.visualizer_handlers[handler.name] = handler

		print("[spb] Add ", handler, " as ", handler.name)


	# --- --- --- --- ---

	def step(self):
		'''1ステップ分の処理を行う'''
		if self.operation==False:
			return

		# Step Enable処理
		enable = bpy.context.scene.spr_step_enabled
		if enable==True or enable==1:
			if not spbapi.GetHapticTimer().IsStarted():
				spbapi.GetHapticTimer().Start()
				if not spb.simulation_started:
					spb.handlers.store("quickstore_on_start")
					spb.simulation_started = True
			if not spbapi.GetPhysicsTimer().IsStarted():
				spbapi.GetPhysicsTimer().Start()
		else:
			if spbapi.GetHapticTimer().IsStarted():
				spbapi.GetHapticTimer().Stop()
				spb.handlers.store("quickstore_on_end")
			if spbapi.GetPhysicsTimer().IsStarted():
				spbapi.GetPhysicsTimer().Stop()

		# キーフレーム書き出しモード
		if self.bake_mode:
			self.bake_step()

		# Minimal Sync
		elif self.is_initialized and bpy.context.scene.spr_minimal_sync:
			self.minimal_step()

		# 通常のSync処理
		else:
			self.full_step()

		# Keyframe Record処理
		if bpy.context.scene.spr_step_enabled and bpy.context.scene.spr_record:
			self.record_step()
		

	def bake_step(self):
		'''キーフレーム書き出し中のステップ処理'''
		self.keys.bake_frame(self.bake_frame_no)
		self.bake_frame_no += 1
		if self.bake_frame_no >= len(self.keys.keyframes):
			self.bake_mode = False
			print("[spb] Bake Complete.")


	def full_step(self):
		'''完全自動同期モード'''
		self.cleanup()
		self.generate()
		self.build()
		self.sync()
		self.handler_step()
		self.store("default", init=True)
		self.is_initialized = True


	def minimal_step(self):
		'''最小同期モード'''
		# 実行時はspr->bpyの最小限同期を実行
		if bpy.context.scene.spr_step_enabled:
			if self.last_step_enabled==False:
				self.full_step()
			else:
				self.extreme_minimal_sync()
				# self.minimal_sync()
				# self.handler_step()

		self.last_step_enabled = bpy.context.scene.spr_step_enabled


	def record_step(self):
		'''キーフレーム記録のステップ処理'''
		if self.record_step_count % bpy.context.scene.spr_record_every_n_steps == 0:
			self.record()
		self.record_step_count += 1
		bpy.context.scene.spr_recording_frame += 1


	# --- --- --- --- ---

	def clear(self):
		'''ハンドラを全削除する'''
		# 関連オブジェクトの破壊
		for handlers in self.handlers_list.__reversed__():
			for handler in handlers.values():
				if handler:
					handler.destroy()

		# ハンドラの全削除
		for handlers in self.handlers_list.__reversed__():
			handlers.clear()

		# シーンハンドラも作り直し
		self.scene_handler = SceneHandler(bpy.context.scene)

		self.is_initialized = False


	# --- --- --- --- ---

	def cleanup(self):
		'''不要になったハンドラを見つけてきて削除する．'''
		for handlers in self.handlers_list.__reversed__():
			self.cleanup_each_handlers(handlers)


	def cleanup_each_handlers(self, handlers):
		remove_list = []

		for handler in handlers.values():
			if (handler is None) or handler.must_removed():
				remove_list.append(handler)

		for handler in remove_list:
			print("[spb] Removing ", handler, " as ", handler.name)
			handlers.pop(handler.name)
			handler.destroy()
			print(" --> fin.")


	# --- --- --- --- ---

	def generate(self):
		'''対応するハンドラ一覧を生成する．
		一度生成済みの場合，未生成のハンドラ(新規追加要素等)のみ生成する．'''
		objects = bpy.data.objects

		# -- Migration Check --
		if bpy.context.scene.spr_need_migration == 0:
			for object in objects:
				if object.spr_physics_type == "Dynamic":
					bpy.context.scene.spr_need_migration = 1

		# -- Solid --
		for object in objects:
			if (not (object.name in self.solid_handlers.keys())) and SolidHandler.is_target(object):
				self.add_handler(SolidHandler(BpyObject(object)))
			if (not (object.name in self.solid_handlers.keys())) and PointerHandler.is_target(object):
				self.add_handler(PointerHandler(BpyObject(object)))

		# ShapeやJointはSolidに依存するので，Solidの生成がひと通り終わってから生成する．
		for object in objects:
			# -- Shape --
			if (not (object.name in self.shape_handlers.keys())) and ShapeHandler.is_target(object):
				self.add_handler(ShapeHandler(BpyObject(object)))

			# -- Joint --
			if (not (object.name in self.joint_handlers.keys())) and JointHandler.is_target(object):
				self.add_handler(JointHandler(BpyObject(object)))

		# LimitはJointに依存するので，Jointの生成がひと通り終わってから生成する．
		for object in objects:
			# -- Limit --
			if (not (object.name in self.limit_handlers.keys())) and LimitHandler.is_target(object):
				self.add_handler(LimitHandler(BpyObject(object)))

		# IK関連はJointやSolidに依存するので，JointやSolidの生成がひと通り終わってから生成する．
		for object in objects:
			# -- IKEff --
			if (not (object.name in self.ikeff_handlers.keys())) and IKEffHandler.is_target(object):
				self.add_handler(IKEffHandler(BpyObject(object)))

			# -- IKAct --
			if (not (object.name in self.ikact_handlers.keys())) and IKActHandler.is_target(object):
				self.add_handler(IKActHandler(BpyObject(object)))

		for group in bpy.data.groups:
			# -- Creature --
			if not (group.name in self.creature_handlers.keys()) and CreatureHandler.is_target(group):
				self.add_handler(CreatureHandler(BpyGroup(group)))

		for group in bpy.data.groups:
			for object in group.objects:
				# -- CRBone --
				if not (object.name in self.crbone_handlers.keys()) and CRBoneHandler.is_target(object):
					self.add_handler(CRBoneHandler(BpyObject(object)))

			for visualizer in group.spr_visualizers:
				# -- Visualizer --
				if not (visualizer.name in self.visualizer_handlers.keys()):
					self.add_handler(VisualizerHandler(BpyGroup(group), visualizer.name))


	def build(self):
		'''各ハンドラのbuild/post_buildを呼び出してSpringhead objectの構築を行う．'''
		# 更新フラグのクリア
		self.scene_handler.updated = False
		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler:
					handler.updated = False

		# 構築・再構築
		if (not self.scene_handler.build_finished):
			self.scene_handler.build()
			self.scene_handler.updated = True

		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler:
					need_rebuild = False
					reasons = []
					# Sprオブジェクト構築情報を更新，変更があればリビルドフラグを立てる
					build_info = deepcopy(handler.build_info)
					handler.update_build_info()
					if (not handler.build_info == build_info):
						need_rebuild = True
						reasons.append("property change")

					# 依存ハンドラがupdatedならリビルドフラグを立てる
					for dh in handler.dependent_handlers:
						if dh.updated:
							need_rebuild = True
							reasons.append(dh.__str__() + " is updated")

					# 本体または関連ハンドラの構築が完了していなければリビルドフラグを立てる（"再"構築ではないかもだけど）
					if (not handler.build_finished):
						need_rebuild = True
						reasons.append("build not completed")
					for dh in handler.dependent_handlers:
						if (not dh.build_finished):
							need_rebuild = True
							reasons.append(dh.__str__() + " build not completed")

					# 未構築またはリビルドフラグが立っている場合，構築を行う
					if (not handler.build_finished) or need_rebuild:
						print(handler.__str__() + " build : ")
						for reason in reasons:
							print("    because : "+reason)
							
						handler.build_finished = False
						handler.build()
						handler.updated = True

		# 後処理
		self.scene_handler.post_build()
		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler and handler.updated and handler.build_finished:
					handler.post_build()

	# --- --- --- --- ---

	def sync(self):
		'''各ハンドラのsyncを呼び出して bpy<=>spr間の同期を行う．'''
		# Syncの順番は他とは異なる（依存関係のため）
		handlers_list = []
		handlers_list.append( self.shape_handlers      )
		handlers_list.append( self.solid_handlers      ) # Inertiaのためshapeよりあと
		handlers_list.append( self.joint_handlers      )
		handlers_list.append( self.limit_handlers      )
		handlers_list.append( self.ikeff_handlers      )
		handlers_list.append( self.ikact_handlers      )
		handlers_list.append( self.creature_handlers   )
		handlers_list.append( self.crbone_handlers     )
		handlers_list.append( self.visualizer_handlers )
		
		# -- Before Sync
		self.scene_handler.before_sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.before_sync()

		# -- Sync
		self.scene_handler.sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.sync()

		# -- After Sync
		self.scene_handler.after_sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.after_sync()


	def extreme_minimal_sync(self):
		'''spr=>bpyの本当に最小限の同期を行う．'''
		handlers_list = []
		handlers_list.append( self.solid_handlers      )
		handlers_list.append( self.creature_handlers   )

		# -- Solid Sync
		for handler in self.solid_handlers.values():
			if handler:
				new_spr = handler.state_sync.get_spr_value_mod(handler.state_sync.spr)
				handler.state_sync.bpy = to_bpy(new_spr)
				handler.state_sync.spr = copy(new_spr)
				handler.state_sync.set_bpy_value(handler.state_sync.bpy)
		
		# -- Scene Step
		self.scene_handler.step()

		# -- Creature Step
		for handler in self.creature_handlers.values():
			if handler:
				handler.step()


	def minimal_sync(self):
		'''spr=>bpyの最小限の同期を行う．'''
		# <!!> 今は使われていない．
		# Extreme Minimal Syncでは不十分な例が出てきたら復活も考慮する．
		# そのうちSync Profile機能が実装されたらそちらに組み入れる予定．

		handlers_list = []
		handlers_list.append( self.solid_handlers      )
		handlers_list.append( self.creature_handlers   )

		# -- Before Sync
		self.scene_handler.before_sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.before_sync()

		# -- Sync
		self.scene_handler.sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.sync_spr_to_bpy()

		# -- After Sync
		self.scene_handler.after_sync()
		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.after_sync()


	# --- --- --- --- ---

	def handler_step(self):
		'''各ハンドラが1ステップごとに実行する内容を実行する．'''
		self.scene_handler.step()

		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler:
					handler.step()
		
	# --- --- --- --- ---

	def store(self, key, init=False):
		'''状態の一時保存を行う．'''
		self.scene_handler.store(key, init)

		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler:
					handler.store(key, init)


	def restore(self, key):
		'''保存しておいた状態に戻す．'''
		self.scene_handler.restore(key)

		for handlers in self.handlers_list:
			for handler in handlers.values():
				if handler:
					handler.restore(key)

	# --- --- --- --- ---

	def record(self):
		'''キーフレームに記録する'''
		frame = bpy.context.scene.spr_recording_frame

		self.scene_handler.record(frame)

		handlers_list = []
		handlers_list.append( self.solid_handlers      )
		handlers_list.append( self.visualizer_handlers )

		for handlers in handlers_list:
			for handler in handlers.values():
				if handler:
					handler.record(frame)
		
		
