# -*- coding: utf-8 -*-

import bpy
import spb
import spb.handlers_imp.joint

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class CRBoneHandler(Handler):
	'''Spr.CRBoneとbpy.objectを同期するクラス．'''

	def __init__(self, object, crBone=None):
		Handler.__init__(self)

		self.object      = object
		self.name        = object.name

		self.crBone      = crBone
		self.creatureHnd = None

		# -- Synchronizers
		self.label_sync = CRBoneLabelSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def update_build_info(self):
		'''Springheadオブジェクトの構築に必要な情報を更新．これが変化すると再構築される'''
		creaturename = ""
		for group in bpy.data.groups:
			if self.name in group.objects:
				creaturename = group.name
		self.build_info["creaturename"] = creaturename

	def build(self):
		'''sprのオブジェクトを構築する'''
		creaturename = self.build_info["creaturename"]

		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# 依存ハンドラのクリア
		# -- CRBoneは構成要素次第で依存ハンドラが変わるので，ここでクリアしておいて出現するたびに追加する．
		self.dependent_handlers = []

		# Find Creature
		self.creatureHnd = spb.handlers.creature_handlers[creaturename]
		if self.creatureHnd is None or self.creatureHnd.crCreature is None:
			print("Error : Could Not Find Creature : ", creaturename)
			return

		if self.creatureHnd.crCreature.NBodies() < 1:
			print("Error : ", creaturename, " does not have body.")
			return

		self.dependent_handlers.append(self.creatureHnd)

		# Create Bone
		crBody = self.creatureHnd.crCreature.GetBody(0)
		self.crBone = crBody.CreateObject(CRBone().GetIfInfoStatic(), CRBoneDesc())

		# Set Name
		self.crBone.SetName("crbone_"+self.object.name)

		# Add Solid
		solidHnd = spb.handlers.solid_handlers[self.name]
		if solidHnd and solidHnd.build_finished:
			self.crBone.AddChildObject(solidHnd.phSolid)
			self.dependent_handlers.append(solidHnd)

		# Add EndEffector
		ikeffHnd = spb.handlers.ikeff_handlers[self.name]
		if ikeffHnd and ikeffHnd.build_finished:
			self.crBone.AddChildObject(ikeffHnd.phIKEff)
			self.dependent_handlers.append(ikeffHnd)

		# <Find Connected Joint>
		joname = None
		for jh in spb.handlers.joint_handlers.values():
			if (not jh.plugHnd is None) and (jh.plugHnd.object.name==self.object.name):
				joname = jh.object.name

		if not joname is None:
			# Add Joint
			jointHnd = spb.handlers.joint_handlers[joname]
			if jointHnd and jointHnd.build_finished:
				self.crBone.AddChildObject(jointHnd.phJoint)
				print("Add", jointHnd.phJoint.GetName(), "to", self.crBone.GetName())
				self.dependent_handlers.append(jointHnd)

			# Add Actuator
			ikactHnd = spb.handlers.ikact_handlers[joname]
			if ikactHnd and ikactHnd.build_finished:
				self.crBone.AddChildObject(ikactHnd.phIKAct)
				# self.crBone.AddChildObject(ikactHnd.jointHnd.phJoint)
				self.dependent_handlers.append(ikactHnd)

			# Joint Handlerの Spring/Damper Syncにmodifierを登録(Spring/Damper Ratio用)
			if ikactHnd and ikactHnd.build_finished:
				jointHnd = ikactHnd.jointHnd
				for sync in jointHnd.synchronizers:
					if type(sync) is spb.handlers_imp.joint.JointSpringSync:
						sync.modifiers.append([self.modify_bpy_joint_spring, self.modify_spr_joint_spring])
					if type(sync) is spb.handlers_imp.joint.JointDamperSync:
						sync.modifiers.append([self.modify_bpy_joint_damper, self.modify_spr_joint_damper])

		# 構築完了フラグを更新
		self.build_finished = (not self.crBone is None)

	def destroy(self):
		if not self.crBone is None:
			# <!!> To Be Implemented
			self.crBone = None
		self.build_finished = False

	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		for ch in spb.handlers.creature_handlers.values():
			if object.name in ch.group.objects and ch.group.spr_creature_group==1:
				return True
		return False

	# --- --- --- --- ---

	# Spring/Damper Ratioを実現するためのmodifier
	def modify_bpy_joint_spring(self, value):
		sr = self.creatureHnd.group.spr_creature_spring_ratio
		return value * sr
	def modify_spr_joint_spring(self, value):
		return value
	def modify_bpy_joint_damper(self, value):
		dr = self.creatureHnd.group.spr_creature_damper_ratio
		return value * dr
	def modify_spr_joint_damper(self, value):
		return value




# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class CRBoneLabelSync(Synchronizer):
	def get_bpy_value(self, last):
		self.handler.object.spr_crbone_label = self.handler.object.spr_CRIKSolid_label # <!!> 移行措置
		return self.handler.object.spr_crbone_label
	def get_spr_value(self, last):
		return self.handler.crBone.GetLabel()
	def set_bpy_value(self, value):
		self.handler.object.spr_crbone_label = value
	def set_spr_value(self, value):
		self.handler.crBone.SetLabel(value)

