# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class IKActHandler(Handler):
	'''Spr.PHIKActuatorとbpy.objectを同期するクラス．'''

	def __init__(self, object, phIKAct=None):
		Handler.__init__(self)

		self.object    = object
		self.name      = object.name

		self.phIKAct   = phIKAct
		self.jointHnd  = None

		# -- Synchronizers
		self.bias_sync = IKActBiasSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def before_sync(self):
		'''Synchronize前に行う処理を記述する．オーバーライド用．'''
		# 各PanelのHold/Applyを読み取って，対応するSynchronizerをHold/Applyする．

		# -- IKActuator Properties
		hold = (self.object.spr_joint_props_hold==1)
		self.bias_sync.hold(hold)

		if (self.object.spr_joint_props_apply==1):
			self.bias_sync.apply()

	def after_sync(self):
		'''Synchronize後に行う処理を記述する．オーバーライド用．'''
		# ApplyボタンをFalseにリセットする．
		self.object.spr_joint_props_apply = 0

	def build(self):
		'''sprのオブジェクトを構築する'''

		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# Get Joint
		self.jointHnd = spb.handlers.joint_handlers[self.name]
		if self.jointHnd is None:
			print("Error: Couldn't Find Joint for IKActuator<!!>", self.name)
			return

		# Create IKAct
		phScene = spbapi.GetFWSdk().GetScene(0).GetPHScene()
		jointtype = self.jointHnd.build_info["jointtype"]
		if jointtype=="Ball":
			ballinfo = PHIKBallActuator().GetIfInfoStatic()
			balldesc = PHIKBallActuatorDesc()
			self.phIKAct = phScene.CreateIKActuator(ballinfo, balldesc)
		elif jointtype=="Hinge":
			hingeinfo = PHIKHingeActuator().GetIfInfoStatic()
			hingedesc = PHIKHingeActuatorDesc()
			self.phIKAct = phScene.CreateIKActuator(hingeinfo, hingedesc)
		else:
			print("Error : IKActuator is only for Ball / Hinge.", self.name)
			return

		self.phIKAct.SetVelocityGain(7) # <!!>
		self.phIKAct.Enable(True)

		# Set Joint
		self.phIKAct.AddChildObject(self.jointHnd.phJoint)
		
		# Set EndEffector
		if self.jointHnd.plugHnd is None:
			print("Error : Joint for IK does not has Plug Object", self.name)
			return

		ikEffHnd = spb.handlers.ikeff_handlers[self.jointHnd.plugHnd.name]
		if (not ikEffHnd is None) and (not ikEffHnd.phIKEff is None):
			self.phIKAct.AddChildObject(ikEffHnd.phIKEff)

		# Set Name
		self.phIKAct.SetName("ika_"+self.object.name)

		# 依存ハンドラのセット
		self.dependent_handlers = [self.jointHnd]

		# 構築完了フラグを更新
		self.build_finished = ( (not self.jointHnd is None) and (not self.phIKAct is None) )

	def post_build(self):
		'''sprオブジェクト構築後に行う必要のある後処理'''
		# 親Actuatorへの登録．全てのActuatorの構築後でなければ実行できない
		# - このIKActに対応するJointのSocket剛体と，
		# - 親のIKActに対応するJointのPlug剛体は一致する  ことを利用して親を探す．
		if (not self.jointHnd.socketHnd is None):
			for iah in spb.handlers.ikact_handlers.values():
				if (not iah.jointHnd.plugHnd is None):
					if (iah.jointHnd.plugHnd.name==self.jointHnd.socketHnd.name):
						if (not iah.phIKAct is None):
							iah.phIKAct.AddChildObject(self.phIKAct)

	def destroy(self):
		if not self.phIKAct is None:
			# とりあえずDisableして転がしとく <!!>
			self.phIKAct.Enable(False)
			self.phIKAct = None
		self.build_finished = False

	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		joint_type = (object.spr_object_type== "Joint")
		ik_enable  = (object.spr_ik_enabled == 1)
		balljoint  = (object.spr_joint_mode == "Ball")
		hingejoint = (object.spr_joint_mode == "Hinge")
		return joint_type and ik_enable and (balljoint or hingejoint)


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class IKActBiasSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_ik_bias
	def get_spr_value(self, last):
		return self.handler.phIKAct.GetBias()
	def set_bpy_value(self, value):
		self.handler.object.spr_ik_bias = value
	def set_spr_value(self, value):
		self.handler.phIKAct.SetBias(value)

