# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class IKEffHandler(Handler):
	'''Spr.PHIKEndEffectorとbpy.objectを同期するクラス．'''

	def __init__(self, object, phIKEff=None):
		Handler.__init__(self)

		self.object    = object
		self.name      = object.name

		self.phIKEff   = phIKEff
		self.solidHnd  = None

		# -- Synchronizers
		self.targetlocalpos_sync = IKEffTargetLocalPosSync(self)
		self.enableposctl_sync   = IKEffEnablePosCtlSync(self)
		self.enableorictl_sync   = IKEffEnableOriCtlSync(self)
		self.targetpose_sync     = IKEffTargetPoseSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def before_sync(self):
		'''Synchronize前に行う処理を記述する．オーバーライド用．'''
		# 各PanelのHold/Applyを読み取って，対応するSynchronizerをHold/Applyする．

		# -- IKEndEffector Properties
		hold = (self.object.spr_solid_props_hold==1)
		self.targetlocalpos_sync.hold(hold)
		self.enableposctl_sync.hold(hold)
		self.enableorictl_sync.hold(hold)

		if (self.object.spr_solid_props_apply==1):
			self.targetlocalpos_sync.apply()
			self.enableposctl_sync.apply()
			self.enableorictl_sync.apply()

		# -- IKEndEffector Control
		self.targetpose_sync.hold( self.object.spr_ik_targetpose_hold==1 )

		if (self.object.spr_ik_targetpose_apply==1):
			self.targetpose_sync.apply()

	def after_sync(self):
		'''Synchronize後に行う処理を記述する．オーバーライド用．'''
		# ApplyボタンをFalseにリセットする．
		self.object.spr_solid_props_apply   = 0
		self.object.spr_ik_targetpose_apply = 0

	def build(self):
		'''sprのオブジェクトを構築する'''
		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# Get Solid
		self.solidHnd = spb.handlers.solid_handlers[self.name]
		if self.solidHnd is None:
			# そのうちちゃんとエラー出力関数作りましょう．同じエラーを無視するとか <!!>
			print("Error: Couldn't Find Solid for IKEndEffector<!!>", self.name)
			return

		# Create IKEff
		phScene = spbapi.GetFWSdk().GetScene(0).GetPHScene()
		self.phIKEff = phScene.CreateIKEndEffector(PHIKEndEffectorDesc())
		self.phIKEff.Enable(True)

		# Set Solid
		self.phIKEff.AddChildObject(self.solidHnd.phSolid)
		
		# Set Name
		self.phIKEff.SetName("ike_"+self.object.name)

		# 依存ハンドラのセット
		self.dependent_handlers = [self.solidHnd]

		# 構築完了フラグを更新
		self.build_finished = ( (not self.solidHnd is None) and (not self.phIKEff is None) )

	def destroy(self):
		if not self.phIKEff is None:
			# とりあえずDisableして転がしとく <!!>
			self.phIKEff.Enable(False)
			self.phIKEff = None
		self.build_finished = False
	
	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		solid_type = (object.spr_object_type=="Solid")
		ik_enable  = (object.spr_ik_enabled == 1)
		return solid_type and ik_enable


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class IKEffTargetLocalPosSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		bpy_tlp = Vector((o.spr_ik_target_local_pos_x,o.spr_ik_target_local_pos_y,o.spr_ik_target_local_pos_z))
		return bpy_tlp
	def get_spr_value(self, last):
		return self.handler.phIKEff.GetTargetLocalPosition()
	def set_bpy_value(self, value):
		self.handler.object.spr_ik_target_local_pos_x = value.x
		self.handler.object.spr_ik_target_local_pos_y = value.y
		self.handler.object.spr_ik_target_local_pos_z = value.z
	def set_spr_value(self, value):
		self.handler.phIKEff.SetTargetLocalPosition(value)

class IKEffEnablePosCtlSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_ik_pos_control_enabled
	def get_spr_value(self, last):
		return self.handler.phIKEff.IsPositionControlEnabled()
	def set_bpy_value(self, value):
		self.handler.object.spr_ik_pos_control_enabled = value
	def set_spr_value(self, value):
		self.handler.phIKEff.EnablePositionControl(value)
	
class IKEffEnableOriCtlSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_ik_ori_control_enabled
	def get_spr_value(self, last):
		return self.handler.phIKEff.IsOrientationControlEnabled()
	def set_bpy_value(self, value):
		self.handler.object.spr_ik_ori_control_enabled = value
	def set_spr_value(self, value):
		self.handler.phIKEff.EnableOrientationControl(value)

class IKEffTargetPoseSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_ik_target_object_enabled==1 and (o.spr_ik_target_object_name in bpy.data.objects):
			# target objectを目標にする場合
			obj_target = bpy.data.objects[o.spr_ik_target_object_name]
			if (obj_target.parent is None and (not obj_target.spr_use_matrix_world==1)):
				# Childでない場合
				loc = obj_target.location
				if (not obj_target.rotation_mode=="QUATERNION"):
					obj_target.rotation_mode = "QUATERNION"
				rot = obj_target.rotation_quaternion
			else:
				# Static Child, または Use Matrix Worldの場合
				mat = obj_target.matrix_world
				loc = mat.to_translation()
				rot = mat.to_quaternion()
		else:
			# 目標を値で指定する場合
			loc = Vector((o.spr_ik_target_pos_x,o.spr_ik_target_pos_y,o.spr_ik_target_pos_z))
			ori = Quaterniond()
			# FromEulerを使うときは座標系の違いに注意 : Blender x,y,z →Springhead z,x,y
			ori.FromEuler(Vec3d(rad(o.spr_ik_target_ori_y),rad(o.spr_ik_target_ori_z),rad(o.spr_ik_target_ori_x)))
			rot = to_bpy(ori)

		return [loc, rot]

	def get_spr_value(self, last):
		pos = self.handler.phIKEff.GetTargetPosition()
		ori = self.handler.phIKEff.GetTargetOrientation()
		return [pos, ori]

	def set_bpy_value(self, value):
		loc, rot = value
		o = self.handler.object
		if o.spr_ik_target_object_enabled==1 and (o.spr_ik_target_object_name in bpy.data.objects):
			# target objectを目標にする場合
			obj_target = bpy.data.objects[o.spr_ik_target_object_name]
			if (obj_target.parent is None and (not obj_target.spr_use_matrix_world==1)):
				# Childでない場合
				# obj_target.location = loc
				# obj_target.rotation_quaternion = rot
				pass
			else:
				# Static Child, または Use Matrix Worldの場合
				# ＜Not Implemented＞ <!!>  というよりStatic Child自体がObsoleted予定なので
				pass
		else:
			# 目標を値で指定する場合
			o.spr_ik_target_pos_x = loc.x
			o.spr_ik_target_pos_y = loc.y
			o.spr_ik_target_pos_z = loc.z

			v = Vec3d()
			to_spr(rot).ToEuler(v)
			# ToEulerを使うときは座標系の違いに注意 : Blender x,y,z →Springhead z,x,y
			o.spr_ik_target_ori_y = deg(v.x)
			o.spr_ik_target_ori_z = deg(v.y)
			o.spr_ik_target_ori_x = deg(v.z)

	def set_spr_value(self, value):
		pos, ori = value
		self.handler.phIKEff.SetTargetPosition(pos)
		self.handler.phIKEff.SetTargetOrientation(ori)
