# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class JointHandler(Handler):
	'''Spr.PHJointとbpy.objectを同期するクラス．'''

	def __init__(self, object, phJoint=None):
		Handler.__init__(self)

		self.object    = object
		self.name      = object.name

		self.phJoint   = phJoint
		self.plugHnd   = None
		self.socketHnd = None

		# -- Synchronizers
		self.collision_sync      = JointCollisionSync(self)
		self.enable_sync         = JointEnableSync(self)
		self.plugsocketpose_sync = JointPlugSocketPoseSync(self)
		self.spring_sync         = JointSpringSync(self)
		self.damper_sync         = JointDamperSync(self)
		self.springori_sync      = JointSpringOriSync(self)
		self.damperori_sync      = JointDamperOriSync(self)
		self.targetposition_sync = JointTargetPositionSync(self)
		self.targetvelocity_sync = JointTargetVelocitySync(self)
		self.offsetforce_sync    = JointOffsetForceSync(self)
		self.maxforce_sync       = JointMaxForceSync(self)


	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def before_sync(self):
		'''Synchronize前に行う処理を記述する．オーバーライド用．'''
		# 各PanelのHold/Applyを読み取って，対応するSynchronizerをHold/Applyする．

		# -- Joint Panel
		hold = (self.object.spr_joint_props_hold==1)
		self.collision_sync.hold(hold)
		self.enable_sync.hold(hold)
		self.plugsocketpose_sync.hold(hold)
		self.spring_sync.hold(hold)
		self.damper_sync.hold(hold)
		self.springori_sync.hold(hold)
		self.damperori_sync.hold(hold)
		self.maxforce_sync.hold(hold)

		if (self.object.spr_joint_props_apply==1):
			self.collision_sync.apply()
			self.enable_sync.apply()
			self.plugsocketpose_sync.apply()
			self.spring_sync.apply()
			self.damper_sync.apply()
			self.springori_sync.apply()
			self.damperori_sync.apply()
			self.maxforce_sync.apply()

		# -- Joint Control Panel
		self.targetposition_sync.hold( self.object.spr_joint_targetposition_hold==1)
		self.targetvelocity_sync.hold( self.object.spr_joint_velocity_hold==1)
		self.offsetforce_sync.hold(    self.object.spr_joint_offsetforce_hold==1)

		if (self.object.spr_joint_targetposition_apply==1):
			self.targetposition_sync.apply()
		if (self.object.spr_joint_velocity_apply==1):
			self.targetvelocity_sync.apply()
		if (self.object.spr_joint_offsetforce_apply==1):
			self.offsetforce_sync.apply()

	def after_sync(self):
		'''Synchronize後に行う処理を記述する．オーバーライド用．'''
		# ApplyボタンをFalseにリセットする．
		self.object.spr_joint_props_apply = 0
		self.object.spr_joint_targetposition_apply = 0
		self.object.spr_joint_velocity_apply       = 0
		self.object.spr_joint_offsetforce_apply    = 0

	def update_build_info(self):
		'''Springheadオブジェクトの構築に必要な情報を更新．これが変化すると再構築される'''
		# 関節の種類
		self.build_info["jointtype"] = self.object.spr_joint_mode

		# Socketの剛体や軸オブジェクトの名前
		if self.object.spr_joint_socket_target == "Solid Name":
			# -- 剛体名をSocketに指定する場合
			self.build_info["sockname"]  = self.object.spr_joint_socket_target_name
			self.build_info["sockaxis"]  = self.object.name
		elif self.object.spr_joint_socket_target == "Axis Name":
			# -- Socket座標系用Emptyオブジェクト名をSocketに指定する場合（Socket剛体はその親）
			sockaxis = self.object.spr_joint_socket_target_name
			if (sockaxis in  bpy.data.objects) and (not bpy.data.objects[sockaxis].parent is None) and (bpy.data.objects[sockaxis].parent.name in spb.handlers.solid_handlers):
				self.build_info["sockname"]  = bpy.data.objects[sockaxis].parent.name
				self.build_info["sockaxis"]  = sockaxis
			else:
				print("Maybe Error : Something Insufficient for Joint Socket <!!>")
				self.build_info["sockname"]  = sockaxis
				self.build_info["sockaxis"]  = sockaxis

		# Plugの剛体や軸オブジェクトの名前
		if self.object.spr_joint_plug_target == "Parent":
			if not self.object.parent is None:
				self.build_info["plugname"]  = self.object.parent.name
			else:
				self.build_info["plugname"]  = ""
			self.build_info["plugaxis"]  = self.object.name

		# Socket剛体やPlug剛体が存在するかどうか
		self.build_info["sockalive"] = (self.build_info["sockname"] in spb.handlers.solid_handlers)
		self.build_info["plugalive"] = (self.build_info["plugname"] in spb.handlers.solid_handlers)

	def build(self):
		'''対応するSpringhead側オブジェクトを構築する．依存ハンドラの取得も行う．'''

		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# Get Socket
		sockname = self.build_info["sockname"]
		self.socketHnd = spb.handlers.solid_handlers[sockname]
		if self.socketHnd is None:
			print("Error: Something Error about Joint Socket  <!!>")
			return

		# Get Plug
		plugname = self.build_info["plugname"]
		self.plugHnd = spb.handlers.solid_handlers[plugname]
		if self.plugHnd is None:
			print("Error: Something Error about Joint Plug  <!!>")
			return

		if not plugname in bpy.data.objects:
			print("Error: Plug Name '" + plugname + "' not found in objects")
			return

		if plugname==self.socketHnd.object.name:
			print("Error: Plug Name '" + plugname + "' is same as socket")
			return
		
		# Create Joint
		jointtype = self.build_info["jointtype"]
		self.phJoint = spbapi.CreateJoint(self.socketHnd.phSolid, self.plugHnd.phSolid, jointtype)

		# Set Name
		self.phJoint.SetName("jo_"+self.object.name)

		# 依存ハンドラのセット
		self.dependent_handlers = [self.socketHnd, self.plugHnd]

		# 構築完了フラグを更新
		self.build_finished = True
		if self.plugHnd is None or self.plugHnd.phSolid is None:
			self.build_finished = False
		if self.socketHnd is None or self.socketHnd.phSolid is None:
			self.build_finished = False
		if self.phJoint is None:
			self.build_finished = False

	def destroy(self):
		'''Springheadオブジェクトを削除する．存在しないものものを消さないように気をつけること'''
		if not self.phJoint is None:
			spbapi.RemoveJoint(self.phJoint)
			self.phJoint = None
		self.build_finished = False

	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		joint_type  = (object.spr_object_type=="Joint")
		return joint_type


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class JointPlugSocketPoseSync(Synchronizer):
	def is_spr_changed(self):
		return False

	def get_bpy_value(self, last):
		sa = self.handler.object
		pa = self.handler.object
		if self.handler.build_info["sockaxis"] in bpy.data.objects:
			sa = bpy.data.objects[self.handler.build_info["sockaxis"]]
		if self.handler.build_info["plugaxis"] in bpy.data.objects:
			pa = bpy.data.objects[self.handler.build_info["plugaxis"]]

		so = self.handler.socketHnd.object
		po = self.handler.plugHnd.object

		return [po.scale, po.matrix_parent_inverse, pa.location, pa.rotation_quaternion, so.scale, so.matrix_parent_inverse, sa.location, sa.rotation_quaternion]

	def set_spr_value(self, value):
		sa = self.handler.object
		pa = self.handler.object
		if self.handler.build_info["sockaxis"] in bpy.data.objects:
			sa = bpy.data.objects[self.handler.build_info["sockaxis"]]
		if self.handler.build_info["plugaxis"] in bpy.data.objects:
			pa = bpy.data.objects[self.handler.build_info["plugaxis"]]

		so = self.handler.socketHnd.object
		po = self.handler.plugHnd.object

		p_rel        = scale_matrix(po.scale) * po.matrix_world.inverted() * pa.matrix_world
		bpy_plug_pos = p_rel.to_translation()
		bpy_plug_ori = p_rel.to_quaternion()

		s_rel        = scale_matrix(so.scale) * so.matrix_world.inverted() * sa.matrix_world
		bpy_sock_pos = s_rel.to_translation()
		bpy_sock_ori = s_rel.to_quaternion()

		pp = bpy_plug_pos; po = bpy_plug_ori
		sp = bpy_sock_pos; so = bpy_sock_ori
		self.handler.phJoint.SetPlugPose(  Posed(Vec3d(pp.x,pp.y,pp.z), Quaterniond(po.w,po.x,po.y,po.z)))
		self.handler.phJoint.SetSocketPose(Posed(Vec3d(sp.x,sp.y,sp.z), Quaterniond(so.w,so.x,so.y,so.z)))


class JointCollisionSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_joint_collision
	def set_spr_value(self, value):
		phScene = spbapi.GetFWSdk().GetScene(0).GetPHScene()
		if value==0:
			phScene.SetContactMode(self.handler.plugHnd.phSolid, self.handler.socketHnd.phSolid, 0)
		else:
			phScene.SetContactMode(self.handler.plugHnd.phSolid, self.handler.socketHnd.phSolid, 2)
		
class JointEnableSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_joint_enabled
	def get_spr_value(self, last):
		return self.handler.phJoint.IsEnabled()
	def set_bpy_value(self, value):
		self.handler.object.spr_joint_enabled = value
	def set_spr_value(self, value):
		self.handler.phJoint.Enable(value)

class JointSpringSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			bpy_spring = Vector((o.spr_spring_x,o.spr_spring_y,o.spr_spring_z))
		else:
			bpy_spring = o.spr_spring
		return bpy_spring
	def get_spr_value(self, last):
		return self.handler.phJoint.GetSpring()
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			o.spr_spring_x = value.x
			o.spr_spring_y = value.y
			o.spr_spring_z = value.z
		else:
			o.spr_spring = value
	def set_spr_value(self, value):
		self.handler.phJoint.SetSpring(value)

class JointDamperSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			bpy_damper = Vector((o.spr_damper_x,o.spr_damper_y,o.spr_damper_z))
		else:
			bpy_damper = o.spr_damper
		return bpy_damper
	def get_spr_value(self, last):
		return self.handler.phJoint.GetDamper()
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			o.spr_damper_x = value.x
			o.spr_damper_y = value.y
			o.spr_damper_z = value.z
		else:
			o.spr_damper = value
	def set_spr_value(self, value):
		self.handler.phJoint.SetDamper(value)

class JointSpringOriSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			return o.spr_spring_ori
		else:
			return None
	def get_spr_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			return self.handler.phJoint.GetSpringOri()
		else:
			return None
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			o.spr_spring_ori = value
	def set_spr_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			self.handler.phJoint.SetSpringOri(value)

class JointDamperOriSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			return o.spr_damper_ori
		else:
			return None
	def get_spr_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			return self.handler.phJoint.GetDamperOri()
		else:
			return None
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			o.spr_damper_ori = value
	def set_spr_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Spring":
			self.handler.phJoint.SetDamperOri(value)

class JointTargetPositionSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge":
			return rad(o.spr_joint_target_position)
		elif o.spr_joint_mode=="Slider":
			return o.spr_joint_target_position
		elif o.spr_joint_mode=="Ball":
			####FromEuler/ToEulerを使うときは座標系の違いに注意 Blender x,y,z →Springhead z,x,y
			tp = Quaterniond()
			tp.FromEuler(Vec3d(rad(o.spr_joint_target_position_y), rad(o.spr_joint_target_position_z), rad(o.spr_joint_target_position_x)))
			return tp
		else:
			return None
	def get_spr_value(self, last):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			return self.handler.phJoint.GetTargetPosition()
		else:
			return None
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge":
			o.spr_joint_target_position = deg(value)
		elif o.spr_joint_mode=="Slider":
			o.spr_joint_target_position = value
		elif o.spr_joint_mode=="Ball":
			####FromEuler/ToEulerを使うときは座標系の違いに注意 Blender x,y,z →Springhead z,x,y
			v = Vec3d()
			to_spr(value).ToEuler(v)
			o.spr_joint_target_position_y = deg(v.x)
			o.spr_joint_target_position_z = deg(v.y)
			o.spr_joint_target_position_x = deg(v.z)
	def set_spr_value(self, value):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			self.handler.phJoint.SetTargetPosition(value)

class JointTargetVelocitySync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge" or o.spr_joint_mode=="Slider":
			return o.spr_joint_target_velocity
		elif o.spr_joint_mode=="Ball":
			return Vec3d(o.spr_joint_target_velocity_x, o.spr_joint_target_velocity_y, o.spr_joint_target_velocity_z)
		else:
			return None
	def get_spr_value(self, last):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			return self.handler.phJoint.GetTargetVelocity()
		else:
			return None
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge" or o.spr_joint_mode=="Slider":
			o.spr_joint_target_velocity = value
		elif o.spr_joint_mode=="Ball":
			o.spr_joint_target_velocity_x = value.x
			o.spr_joint_target_velocity_y = value.y
			o.spr_joint_target_velocity_z = value.z
	def set_spr_value(self, value):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			self.handler.phJoint.SetTargetVelocity(value)

class JointOffsetForceSync(Synchronizer):
	def get_bpy_value(self, last):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge" or o.spr_joint_mode=="Slider":
			return o.spr_joint_offset_force
		elif o.spr_joint_mode=="Ball":
			return Vec3d(o.spr_joint_offset_force_x, o.spr_joint_offset_force_y, o.spr_joint_offset_force_z)
		else:
			return None
	def get_spr_value(self, last):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			return self.handler.phJoint.GetOffsetForce()
		else:
			return None
	def set_bpy_value(self, value):
		o = self.handler.object
		if o.spr_joint_mode=="Hinge" or o.spr_joint_mode=="Slider":
			o.spr_joint_offset_force = value
		elif o.spr_joint_mode=="Ball":
			o.spr_joint_offset_force_x = value.x
			o.spr_joint_offset_force_y = value.y
			o.spr_joint_offset_force_z = value.z
	def set_spr_value(self, value):
		o = self.handler.object
		if not o.spr_joint_mode=="Spring":
			self.handler.phJoint.SetOffsetForce(value)

class JointMaxForceSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_joint_max_force
	def get_spr_value(self, last):
		return self.handler.phJoint.GetMaxForce()
	def set_bpy_value(self, value):
		self.handler.object.spr_joint_max_force = value
	def set_spr_value(self, value):
		self.handler.phJoint.SetMaxForce(value)







# --- --- --- --- --- --- --- --- --- --- 
# ボタンを押した際の動作

# 関節のパラメータApplyボタンを押した際の動作
class OBJECT_OT_SprJointParametersApply(bpy.types.Operator):
	bl_idname = "spr.joint_parameters_apply"
	bl_label = "[spr]Apply Joint Parameters"
	bl_description = "[Spr]"
	
	def execute(self,context):
		joint_handler = spb.handlers.joint_handlers[context.object.name]
		if joint_handler:
			joint_handler.apply_bpy_change = True
			joint_handler.sync()
		
		return{'FINISHED'}

###plug handle
#####memo: panelのclassではオブジェクトの変更は基本的に許されない -> 外部やボタンのclassならOK
class OBJECT_OT_SprPlugHandleOn(bpy.types.Operator): 
	bl_idname = "spr.plug_handle_on"
	bl_label = "[spr]joint plug handle ON"
	bl_description = "[Spr]"
	
	def execute(self,context):
		
		obj = context.object
		scene = context.scene
		
		obj_plug_handle = None
		
		try:#既存なら既存のものを使う
			obj_plug_handle = bpy.data.objects["p_"+obj.name]
			#Crear Parent Inverse
			bpy.ops.object.select_all(action='DESELECT')
			obj_plug_handle.hide_select = False
			obj_plug_handle.select = True
			bpy.ops.object.parent_clear(type="CLEAR_INVERSE")
			
		except:#なければ作成
			obj_plug_handle = bpy.data.objects.new("p_"+obj.name, None)
			scene.objects.link(obj_plug_handle)
			obj_plug_handle.empty_draw_type = "ARROWS"
			obj_plug_handle.show_x_ray = True
			obj_plug_handle.parent = obj
			bpy.ops.object.select_all(action='DESELECT')
			obj_plug_handle.select = True
			
		#位置姿勢をデータに合わせる
		obj_plug_handle.location = (obj.spr_joint_plug_pos_x, obj.spr_joint_plug_pos_y, obj.spr_joint_plug_pos_z)
		obj_plug_handle.rotation_mode = "QUATERNION"
		obj_plug_handle.rotation_quaternion = Quaternion((obj.spr_joint_plug_ori_w, obj.spr_joint_plug_ori_x, obj.spr_joint_plug_ori_y, obj.spr_joint_plug_ori_z))
		
		#handle modeフラグ
		obj.spr_joint_plug_handle = 1
		
		return{'FINISHED'}

class OBJECT_OT_SprPlugHandleApply(bpy.types.Operator): 
	bl_idname = "spr.plug_handle_apply"
	bl_label = "[spr]joint plug handle Apply"
	bl_description = "[Spr]"
	
	def execute(self,context):
		
		obj = context.object
		scene = context.scene
		
		obj_plug_handle = None
		
		try:#既存のオブジェクトを取得
			obj_plug_handle = bpy.data.objects["p_"+obj.name]
		except:
			#handle modeフラグ
			obj.spr_joint_plug_handle = 0
			return{'FINISHED'}
		
		#Apply
		loc = obj_plug_handle.location
		rot = obj_plug_handle.rotation_quaternion#(w,x,y,z)
		
		obj.spr_joint_plug_pos_x = loc.x * obj.scale.x
		obj.spr_joint_plug_pos_y = loc.y * obj.scale.y
		obj.spr_joint_plug_pos_z = loc.z * obj.scale.z
		
		obj.spr_joint_plug_ori_w = rot.w
		obj.spr_joint_plug_ori_x = rot.x
		obj.spr_joint_plug_ori_y = rot.y
		obj.spr_joint_plug_ori_z = rot.z
		
		#選択不能にする(削除はしない)
		obj_plug_handle.hide_select = True
		
		#handle modeフラグ
		obj.spr_joint_plug_handle = 0
		
		return{'FINISHED'}

###socket handle
####memo: panelのclassではオブジェクトの変更は基本的に許されない -> 外部やボタンのclassならOK
class OBJECT_OT_SprSocketHandleOn(bpy.types.Operator): 
	bl_idname = "spr.socket_handle_on"
	bl_label = "[spr]joint socket handle ON"
	bl_description = "[Spr]"
	
	def execute(self,context):
		
		obj = context.object
		scene = context.scene
		
		obj_socket_handle = None
		
		try:#既存なら既存のものを使う
			obj_socket_handle = bpy.data.objects["s_"+obj.name]
			#Crear Parent Inverse
			bpy.ops.object.select_all(action='DESELECT')
			obj_socket_handle.hide_select = False
			obj_socket_handle.select = True
			bpy.ops.object.parent_clear(type="CLEAR_INVERSE")
			
		except:#なければ作成
			obj_socket_handle = bpy.data.objects.new("s_"+obj.name, None)
			scene.objects.link(obj_socket_handle)
			obj_socket_handle.empty_draw_type = "ARROWS"
			bpy.ops.object.select_all(action='DESELECT')
			obj_socket_handle.select = True

		
		#joint親オブジェクトにparent
		obj_patent = None
		try:#joint親オブジェクト取得
			obj_parent = bpy.data.objects[obj.spr_joint_target]
			obj_socket_handle.parent = obj_parent
		except:
			#print("Error: Joint Target Object is Missing")
			return{'FINISHED'}
		
		#位置姿勢をデータに合わせる
		obj_socket_handle.location = (obj.spr_joint_socket_pos_x, obj.spr_joint_socket_pos_y, obj.spr_joint_socket_pos_z)
		obj_socket_handle.rotation_mode = "QUATERNION"
		obj_socket_handle.rotation_quaternion = Quaternion((obj.spr_joint_socket_ori_w, obj.spr_joint_socket_ori_x, obj.spr_joint_socket_ori_y, obj.spr_joint_socket_ori_z))
		
		#handle modeフラグ
		obj.spr_joint_socket_handle = 1
		
		return{'FINISHED'}

class OBJECT_OT_SprSocketHandleApply(bpy.types.Operator): 
	bl_idname = "spr.socket_handle_apply"
	bl_label = "[spr]joint socket handle Apply"
	bl_description = "[Spr]"
	
	def execute(self,context):
		
		obj = context.object
		scene = context.scene
		
		obj_socket_handle = None
		
		try:#既存のオブジェクトを取得
			obj_socket_handle = bpy.data.objects["s_"+obj.name]
		except:
			#handle modeフラグ
			obj.spr_joint_socket_handle = 0
			return{'FINISHED'}
		
		#Apply
		loc = obj_socket_handle.location
		rot = obj_socket_handle.rotation_quaternion#(w,x,y,z)
		
		obj.spr_joint_socket_pos_x = loc.x
		obj.spr_joint_socket_pos_y = loc.y
		obj.spr_joint_socket_pos_z = loc.z
		
		obj.spr_joint_socket_ori_w = rot.w
		obj.spr_joint_socket_ori_x = rot.x
		obj.spr_joint_socket_ori_y = rot.y
		obj.spr_joint_socket_ori_z = rot.z
		
		#選択不能にする(削除はしない)
		obj_socket_handle.hide_select = True
		
		#handle modeフラグ
		obj.spr_joint_socket_handle = 0
		
		return{'FINISHED'}

###Limit Infinit (limit解除ボタン)
class OBJECT_OT_SprLimitInfRangeMax(bpy.types.Operator):
	bl_idname = "spr.limit_inf_range_max"
	bl_label = "[spr]Limit Inf Range Max"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_range_max = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprLimitInfRangeMin(bpy.types.Operator):
	bl_idname = "spr.limit_inf_range_min"
	bl_label = "[spr]Limit Inf Range min"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_range_min = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprLimitInfSwingMax(bpy.types.Operator):
	bl_idname = "spr.limit_inf_swing_max"
	bl_label = "[spr]Limit Inf Swing Max"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_swing_max = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprLimitInfSwingMin(bpy.types.Operator):
	bl_idname = "spr.limit_inf_swing_min"
	bl_label = "[spr]Limit Inf Swing min"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_swing_min = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprLimitInfTwistMax(bpy.types.Operator):
	bl_idname = "spr.limit_inf_twist_max"
	bl_label = "[spr]Limit Inf twist Max"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_twist_max = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprLimitInfTwistMin(bpy.types.Operator):
	bl_idname = "spr.limit_inf_twist_min"
	bl_label = "[spr]Limit Inf twist min"
	bl_description = "[Spr]"
	
	def execute(self,context):
		obj = context.object
		obj.spr_limit_twist_min = 3e+39
		
		return{'FINISHED'}

class OBJECT_OT_SprAddJoint(bpy.types.Operator):
	bl_idname = "spr.add_joint"
	bl_label = "[spr]Add Joint"
	bl_description = "[Spr]Add Joint as Empty, Child of This Object"
	
	def execute(self,context):
		obj = context.object
		scene = context.scene
		
		obj_joint = None
		
		i = 0
		while 1:
			if "jo"+str(i)+obj.name in bpy.data.objects:
				i += 1
			else:
				obj_joint = bpy.data.objects.new("jo"+str(i)+obj.name, None)
				scene.objects.link(obj_joint)
				obj_joint.empty_draw_type = "ARROWS"
				obj_joint.parent = obj
				obj_joint.rotation_mode = "QUATERNION"
				
				obj_joint.spr_additional_joint = 1
				obj_joint.spr_joint_mode = obj.spr_joint_mode
				
				
				break
		
		return{'FINISHED'}
