# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class LimitHandler(Handler):
	'''Spr.PHJointLimitとbpy.objectを同期するクラス．'''

	def __init__(self, object, phLimit=None):
		Handler.__init__(self)

		self.object    = object
		self.name      = object.name

		self.phLimit   = phLimit
		self.jointHnd  = None

		# -- Synchronizers
		self.spring_sync = LimitSpringSync(self)
		self.damper_sync = LimitDamperSync(self)
		self.swing_sync  = LimitSwingSync(self)
		self.twist_sync  = LimitTwistSync(self)
		self.range_sync  = LimitRangeSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド
	
	def update_build_info(self):
		'''Springheadオブジェクトの構築に必要な情報を更新．これが変化すると再構築される'''
		# 親階層を辿っていって最後に見つかったJointHandlerを，Limit適用先とする
		o = self.object
		target = o
		while (not o.parent is None):
			o = o.parent
			if not spb.handlers.joint_handlers[o.name] is None:
				target = o
		self.build_info["jointname"] = target.name

	def before_sync(self):
		'''Synchronize前に行う処理を記述する．オーバーライド用．'''
		# 各PanelのHold/Applyを読み取って，対応するSynchronizerをHold/Applyする．

		# -- Joint Limit
		hold = (self.object.spr_joint_props_hold==1)
		self.spring_sync.hold(hold)
		self.damper_sync.hold(hold)
		self.swing_sync.hold(hold)
		self.twist_sync.hold(hold)
		self.range_sync.hold(hold)

		if (self.object.spr_joint_props_apply==1):
			self.spring_sync.apply()
			self.damper_sync.apply()
			self.swing_sync.apply()
			self.twist_sync.apply()
			self.range_sync.apply()

	def after_sync(self):
		'''Synchronize後に行う処理を記述する．オーバーライド用．'''
		# ApplyボタンをFalseにリセットする．
		self.object.spr_joint_props_apply = 0

	def build(self):
		'''sprのオブジェクトを構築する'''

		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# Get Joint
		jointname = self.build_info["jointname"]
		self.jointHnd = spb.handlers.joint_handlers[jointname]
		if self.jointHnd is None:
			print("Error: Something Error about Joint Socket  <!!>")
			return
		
		# Create Joint Limit
		jointtype = self.jointHnd.build_info["jointtype"]
		if jointtype=="Hinge":
			self.phLimit = self.jointHnd.phJoint.CreateLimit(PH1DJointLimitDesc())
		elif jointtype=="Ball":
			self.phLimit = self.jointHnd.phJoint.CreateLimit(PHBallJointConeLimit.GetIfInfoStatic(), PHBallJointConeLimitDesc())

		# 依存ハンドラのセット
		self.dependent_handlers = [self.jointHnd]

		# 構築完了フラグを更新
		self.build_finished = ( (not self.jointHnd is None) and (not self.phLimit is None) )

	def destroy(self):
		if (not self.phLimit is None) and (not self.jointHnd is None):
			self.jointHnd.phJoint.DelChildObject(self.phLimit)
			self.phLimit = None
		self.build_finished = False

	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		# 親階層を辿っていって最後に見つかったJointHandlerを，Limit適用先とする
		o = object
		target = o
		while (not o.parent is None):
			o = o.parent
			if not spb.handlers.joint_handlers[o.name] is None:
				target = o
		jointtype = target.spr_joint_mode

		if object.spr_joint_limit_enable==1:
			return True

		return False



# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

def infrad(v):
	if v > 3e+37:
		return 3e+39
	else:
		return rad(v)

def infdeg(v):
	if v > 3e+37:
		return 3e+39
	else:
		return deg(v)

def inf(v):
	if v > 3e+37:
		return 3e+39
	else:
		return v

def cmpinf(v1, v2):
	return ((v1.x > 3e+37 and v2.x > 3e+37) or (v1.x == v2.x)) and ((v1.y > 3e+37 and v2.y > 3e+37) or (v1.y == v2.y))

class LimitSpringSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_limit_spring
	def get_spr_value(self, last):
		return self.handler.phLimit.GetSpring()
	def set_bpy_value(self, value):
		self.handler.object.spr_limit_spring = value
	def set_spr_value(self, value):
		self.handler.phLimit.SetSpring(value)

class LimitDamperSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.spr_limit_damper
	def get_spr_value(self, last):
		return self.handler.phLimit.GetDamper()
	def set_bpy_value(self, value):
		self.handler.object.spr_limit_damper = value
	def set_spr_value(self, value):
		self.handler.phLimit.SetDamper(value)

class LimitSwingSync(Synchronizer):
	def is_bpy_changed(self):
		'''bpyの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.bpy, self.get_bpy_value_mod(self.bpy)) )
	def is_spr_changed(self):
		'''sprの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.spr, self.get_spr_value_mod(self.spr)) )
	def get_bpy_value(self, last):
		o  = self.handler.object
		return Vec2d(infrad(o.spr_limit_swing_min), infrad(o.spr_limit_swing_max))
	def get_spr_value(self, last):
		o  = self.handler.object
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Ball":
			r = Vec2d()
			self.handler.phLimit.GetSwingRange(r)
			return r
		else:
			return self.bpy
	def set_bpy_value(self, value):
		o = self.handler.object
		o.spr_limit_swing_min = infdeg(value.x)
		o.spr_limit_swing_max = infdeg(value.y)
	def set_spr_value(self, value):
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Ball":
			self.handler.phLimit.SetSwingRange(value)

class LimitTwistSync(Synchronizer):
	def is_bpy_changed(self):
		'''bpyの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.bpy, self.get_bpy_value_mod(self.bpy)) )
	def is_spr_changed(self):
		'''sprの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.spr, self.get_spr_value_mod(self.spr)) )
	def get_bpy_value(self, last):
		o  = self.handler.object
		return Vec2d(infrad(o.spr_limit_twist_min), infrad(o.spr_limit_twist_max))
	def get_spr_value(self, last):
		o  = self.handler.object
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Ball":
			r = Vec2d()
			self.handler.phLimit.GetTwistRange(r)
			return r
		else:
			return self.bpy
	def set_bpy_value(self, value):
		o = self.handler.object
		o.spr_limit_twist_min = infdeg(value.x)
		o.spr_limit_twist_max = infdeg(value.y)
	def set_spr_value(self, value):
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Ball":
			self.handler.phLimit.SetTwistRange(value)

class LimitRangeSync(Synchronizer):
	def is_bpy_changed(self):
		'''bpyの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.bpy, self.get_bpy_value_mod(self.bpy)) )
	def is_spr_changed(self):
		'''sprの新しい値を取得して，変更があったかを返す'''
		return( not cmpinf(self.spr, self.get_spr_value_mod(self.spr)) )
	def get_bpy_value(self, last):
		o  = self.handler.object
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Hinge":
			return Vec2d(infrad(o.spr_limit_range_min), infrad(o.spr_limit_range_max))
		else:
			return Vec2d(inf(o.spr_limit_range_min), inf(o.spr_limit_range_max))
	def get_spr_value(self, last):
		o  = self.handler.object
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Hinge":
			r = Vec2d()
			self.handler.phLimit.GetRange(r)
			return r
		else:
			return self.bpy
	def set_bpy_value(self, value):
		o = self.handler.object
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Hinge":
			o.spr_limit_range_min = infdeg(value.x)
			o.spr_limit_range_max = infdeg(value.y)
		else:
			o.spr_limit_range_min = inf(value.x)
			o.spr_limit_range_max = inf(value.y)
	def set_spr_value(self, value):
		jo = self.handler.jointHnd.object
		if jo.spr_joint_mode=="Hinge" or jo.spr_joint_mode=="Slider":
			self.handler.phLimit.SetRange(value)

