# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *
from   spb.scriptengine import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class SceneHandler(Handler):
	'''Spr.PHSceneとbpy.sceneを同期するクラス．'''

	def __init__(self, scene, fwScene=None):
		Handler.__init__(self)

		self.fwScene    = fwScene
		self.name       = "scene"

		self.rule       = ScriptEngine()
		self.rule.set_name("spr_script")
		self.first_step = True

		# -- Synchronizers
		self.gravity_sync            = SceneGravitySync(self)
		self.numiteration_sync       = SceneNumIterationSync(self)
		self.timestep_sync           = SceneTimeStepSync(self)
		self.impactthreshold_sync    = SceneImpactThresholdSync(self)
		self.frictionthreshold_sync  = SceneFrictionThresholdSync(self)
		self.contacttolerance_sync   = SceneContactToleranceSync(self)
		self.maxvelocity_sync        = SceneMaxVelocitySync(self)
		self.maxangularvelocity_sync = SceneMaxAngularVelocitySync(self)
		self.ikisenabled_sync        = SceneIKIsEnabledSync(self)
		self.iknumiter_sync          = SceneIKNumIterSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def build(self):
		# spbapi.Init()
		self.fwScene  = spbapi.GetFWSdk().GetScene(0)

		# 構築完了フラグを更新
		self.build_finished = True
		if (spbapi.GetFWSdk().NScene() == 0):
			self.build_finished = False
		else:
			self.fwScene  = spbapi.GetFWSdk().GetScene(0)
			self.build_finished = (not self.fwScene is None)

	def step(self):
		'''1ステップごとに実行する内容を記述する．Ruleの実行など'''
		if bpy.context.scene.spr_run_scripts_enabled==1 and bpy.context.scene.spr_step_enabled==1:
			if self.first_step:
				self.rule.update()
				self.first_step = False
			self.rule.step()

	def record(self, frame):
		'''キーフレームへの記録を行う．'''
		pass
		# <!!> Virtual Baby用の一時コード
		# objects = [
		# 		bpy.data.objects['baby_Armature'].pose.bones['upper_jaw'],
		# 		bpy.data.objects['baby_Armature'].pose.bones['lower_jaw'],
		# 		]
		# for obj in objects:
		# 	spb.handlers.keys.insert(obj, frame, location=True, rotation=True)
		# 	# obj.keyframe_insert(data_path="location", frame=frame)
		# 	# if (not obj.rotation_mode=="QUATERNION"):
		# 	# 	obj.rotation_mode = "QUATERNION"
		# 	# obj.keyframe_insert(data_path="rotation_quaternion", frame=frame)


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class SceneGravitySync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.gravity
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetGravity()
	def set_bpy_value(self, value):
		bpy.context.scene.gravity = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetGravity(value)

class SceneNumIterationSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_num_iteration
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetNumIteration()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_num_iteration = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetNumIteration(value)

class SceneTimeStepSync(Synchronizer):
	def is_bpy_changed(self):
		return (abs(self.bpy - self.get_bpy_value_mod(self.bpy)) > 1e-4) if (not self.bpy is None) else True
	def is_spr_changed(self):
		return (abs(self.spr - self.get_spr_value_mod(self.bpy)) > 1e-4) if (not self.spr is None) else True
	def get_bpy_value(self, last):
		return 1/( bpy.context.scene.game_settings.frequency )
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetTimeStep()
	def set_bpy_value(self, value):
		bpy.context.scene.game_settings.frequency = 1/value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetTimeStep(value)

class SceneImpactThresholdSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_impact_threshold
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetImpactThreshold()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_impact_threshold = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetImpactThreshold(value)

class SceneFrictionThresholdSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_friction_threshold
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetFrictionThreshold()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_friction_threshold = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetFrictionThreshold(value)

class SceneContactToleranceSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_contact_tolerance
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetContactTolerance()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_contact_tolerance = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetContactTolerance(value)

class SceneMaxVelocitySync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_max_velocity
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetMaxVelocity()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_max_velocity = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetMaxVelocity(value)

class SceneMaxAngularVelocitySync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_max_angular_velocity
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetMaxAngularVelocity()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_max_angular_velocity = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().SetMaxAngularVelocity(value)

class SceneIKIsEnabledSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_ik_engine_enabled
	def get_spr_value(self, last):
		return self.handler.fwScene.GetPHScene().GetIKEngine().IsEnabled()
	def set_bpy_value(self, value):
		bpy.context.scene.spr_ik_engine_enabled = value
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().GetIKEngine().Enable(value)

class SceneIKNumIterSync(Synchronizer):
	def get_bpy_value(self, last):
		return bpy.context.scene.spr_ik_iteration_num
	def set_spr_value(self, value):
		self.handler.fwScene.GetPHScene().GetIKEngine().SetNumIter(value)


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# SprSceneパネルのレイアウト
#

class SprScenePanel(bpy.types.Panel):
	bl_label       = "Springhead Scene Settings"
	bl_space_type  = "PROPERTIES"
	bl_region_type = "WINDOW"
	bl_context     = "scene"

	def draw(self, context):
		layout = self.layout
		scene = context.scene
		
		layout.prop(scene, "gravity")
		
		split = layout.split()
		split.label(text="Physics Frame Rate")
		split.prop(scene.game_settings, "frequency", text="FPS", slider=False)

		split = layout.split()
		split.label(text="LCP Num Iteration (default=15)")
		split.prop(scene, "spr_num_iteration", text="NumIteration", slider=False)
		
		split = layout.split()
		split.label(text="IK Iteration Per Step (default=1)")
		split.prop(scene, "spr_ik_iteration_num", text = "IK Iteration(per step)")

		split = layout.split()
		split.label(text="Impact Threshold (default=10.0)")
		split.prop(scene, "spr_impact_threshold", text="ImpactThreshold")
		
		split = layout.split()
		split.label(text="Friction Threshold (default=0.01)")
		split.prop(scene, "spr_friction_threshold", text="FrictionThreshold")
		
		split = layout.split()
		split.label(text="Contact Tolerance (default=0.002)")
		split.prop(scene, "spr_contact_tolerance", text="ContactTolerance")
		
		split = layout.split()
		split.label(text="Max Velocity (default=Inf)")
		split.prop(scene, "spr_max_velocity", text="MaxVelocity")
		
		split = layout.split()
		split.label(text="Max Angular Velocity (default=100)")
		split.prop(scene, "spr_max_angular_velocity", text="MaxAngularVelocity")


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# AddメニューをオーバーライドしてSpringhead ObjectをAddできるようにする

import bl_ui.space_info
from bpy.types import Menu
info_mt_add_draw = bl_ui.space_info.INFO_MT_add.draw
class INFO_MT_add(Menu):
	bl_label = "Add"

	def draw(self, context):
		info_mt_add_draw(self, context)
		layout = self.layout
		layout.separator()

		layout.menu("INFO_MT_spr_add_solid", icon='OUTLINER_OB_MESH')
		layout.menu("INFO_MT_spr_add_joint", icon='CONSTRAINT')

# -- -- -- -- --
class INFO_MT_spr_add_solid(Menu):
	bl_idname = "INFO_MT_spr_add_solid"
	bl_label = "Spr Solid"

	def draw(self, context):
		layout = self.layout

		layout.operator_context = 'EXEC_REGION_WIN'
		layout.operator("spr.add_solid_box",    icon='MESH_CUBE',     text="Box")
		layout.operator("spr.add_solid_sphere", icon='MESH_UVSPHERE', text="Sphere")

class INFO_MT_SprAddSolidBox(bpy.types.Operator):
	bl_idname = "spr.add_solid_box"
	bl_label = "[spr]Box"
	bl_description = "[Spr]Box"
	
	def execute(self,context):
		bpy.ops.mesh.primitive_cube_add()
		obj = bpy.context.object
		obj.spr_object_type = "Solid"
		obj.spr_shape_type	= "Box"
		
		return{'FINISHED'}

class INFO_MT_SprAddSolidSphere(bpy.types.Operator):
	bl_idname = "spr.add_solid_sphere"
	bl_label = "[spr]Sphere Solid"
	bl_description = "[Spr]Sphere Solid"
	
	def execute(self,context):
		bpy.ops.mesh.primitive_uv_sphere_add()
		obj = bpy.context.object
		obj.spr_object_type = "Solid"
		obj.spr_shape_type	= "Sphere"
		
		return{'FINISHED'}

# -- -- -- -- --
class INFO_MT_spr_add_joint(Menu):
	bl_idname = "INFO_MT_spr_add_joint"
	bl_label = "Spr Joint"

	def draw(self, context):
		layout = self.layout

		layout.operator_context = 'EXEC_REGION_WIN'
		layout.operator("spr.add_joint_hinge",  text="Hinge")
		layout.operator("spr.add_joint_slider", text="Slider")
		layout.operator("spr.add_joint_ball",   text="Ball")
		layout.operator("spr.add_joint_spring", text="Spring")

class INFO_MT_SprAddJointHinge(bpy.types.Operator):
	bl_idname = "spr.add_joint_hinge"
	bl_label = "[spr]Hinge Joint"
	bl_description = "[Spr]Hinge Joint"
	
	def execute(self,context):
		bpy.ops.object.empty_add(type='SINGLE_ARROW')
		obj = bpy.context.object
		obj.show_x_ray = True
		obj.spr_object_type = "Joint"
		obj.spr_joint_mode  = "Hinge"
		
		return{'FINISHED'}

class INFO_MT_SprAddJointSlider(bpy.types.Operator):
	bl_idname = "spr.add_joint_slider"
	bl_label = "[spr]Slider Joint"
	bl_description = "[Spr]Slider Joint"
	
	def execute(self,context):
		bpy.ops.object.empty_add(type='SINGLE_ARROW')
		obj = bpy.context.object
		obj.show_x_ray = True
		obj.spr_object_type = "Joint"
		obj.spr_joint_mode  = "Slider"
		
		return{'FINISHED'}

class INFO_MT_SprAddJointBall(bpy.types.Operator):
	bl_idname = "spr.add_joint_ball"
	bl_label = "[spr]Ball Joint"
	bl_description = "[Spr]Ball Joint"
	
	def execute(self,context):
		bpy.ops.object.empty_add(type='SINGLE_ARROW')
		obj = bpy.context.object
		obj.show_x_ray = True
		obj.spr_object_type = "Joint"
		obj.spr_joint_mode  = "Ball"
		
		return{'FINISHED'}

class INFO_MT_SprAddJointSpring(bpy.types.Operator):
	bl_idname = "spr.add_joint_spring"
	bl_label = "[spr]Spring Joint"
	bl_description = "[Spr]Spring Joint"
	
	def execute(self,context):
		bpy.ops.object.empty_add(type='ARROWS')
		obj = bpy.context.object
		obj.show_x_ray = True
		obj.spr_object_type = "Joint"
		obj.spr_joint_mode  = "Spring"
		
		return{'FINISHED'}


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
# File -> Exportメニューをオーバーライドして.sprファイルを出力できるようにする

info_mt_file_export_draw = bl_ui.space_info.INFO_MT_file_export.draw
class INFO_MT_file_export(Menu):
	bl_idname = "INFO_MT_file_export"
	bl_label = "Export"

	def draw(self, context):
		info_mt_file_export_draw(self, context)
		self.layout.operator("spr.export_spr_file", text="Springhead (.spr)")

class INFO_MT_SprExportSprFile(bpy.types.Operator):
	bl_idname = "spr.export_spr_file"
	bl_label = "[spr]Export Springhead File"
	bl_description = "[Spr]Export Springhead File"
	
	def execute(self,context):
		spb.spbapi.SaveSpr(bpy.data.filepath+".spr")
		return{'FINISHED'}

