# -*- coding: utf-8 -*-

import bpy
import spb

from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class ShapeHandler(Handler):
	'''Spr.CDShapeとbpy.objectを同期するクラス．'''

	def __init__(self, object, cdShape=None):
		Handler.__init__(self)

		self.object   = object
		self.name     = object.name

		self.cdShape  = cdShape
		self.solidHnd = None

		# -- Synchronizers
		self.pose_sync           = ShapePoseSync(self)
		self.dimension_sync      = ShapeDimensionSync(self)
		self.elasticity_sync     = ShapeElasticitySync(self)
		self.friction_sync       = ShapeFrictionSync(self)
		self.staticfriction_sync = ShapeStaticFrictionSync(self)

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def before_sync(self):
		'''Synchronize前に行う処理を記述する．オーバーライド用．'''
		# 各PanelのHold/Applyを読み取って，対応するSynchronizerをHold/Applyする．

		# -- Solid Panel
		hold = (self.object.spr_solid_props_hold==1)
		self.elasticity_sync.hold(hold)
		self.friction_sync.hold(hold)
		self.staticfriction_sync.hold(hold)

		if (self.object.spr_solid_props_apply==1):
			self.elasticity_sync.apply()
			self.friction_sync.apply()
			self.staticfriction_sync.apply()

	def after_sync(self):
		'''Synchronize後に行う処理を記述する．オーバーライド用．'''
		# ApplyボタンをFalseにリセットする．
		self.object.spr_solid_props_apply       = 0

	def update_build_info(self):
		'''Springheadオブジェクトの構築に必要な情報を更新．これが変化すると再構築される'''
		# 親階層を辿っていって最後に見つかったSolidHandlerを，形状適用先とする
		o = self.object
		target = o
		while (not o.parent is None):
			o = o.parent
			if not spb.handlers.solid_handlers[o.name] is None:
				target = o
		self.build_info["solidname"] = target.name

		# 形状の種類
		self.build_info["shapetype"] = self.object.spr_shape_type

	def build(self):
		'''対応するSpringhead側オブジェクトを構築する．依存ハンドラの取得も行う．'''
		
		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# Get Solid
		solidname = self.build_info["solidname"]
		self.solidHnd = spb.handlers.solid_handlers[solidname]
		if self.solidHnd is None:
			print("Error : Error on Getting Target Solid for Shape <!!> : ", solidname)
			return

		# Create Shape
		if self.build_info["shapetype"] == "Sphere":
			sphereDesc = CDSphereDesc()
			dim = self.object.dimensions
			radius = (dim.x + dim.y + dim.z) / 3 / 2
			sphereDesc.radius = float(radius)
			self.cdShape = spbapi.GetFWSdk().GetPHSdk().CreateShape(CDSphere.GetIfInfoStatic(), sphereDesc)

		if self.build_info["shapetype"] == "Box":
			boxDesc = CDBoxDesc()
			dim = self.object.dimensions
			boxDesc.boxsize = Vec3d(dim.x, dim.y, dim.z)
			self.cdShape = spbapi.GetFWSdk().GetPHSdk().CreateShape(CDBox.GetIfInfoStatic(), boxDesc)

		if self.build_info["shapetype"] == "Convex":
			convexDesc = CDConvexMeshDesc()
			v_vec3d = []
			for v in self.object.data.vertices:
				scale = self.object.scale
				v_vec3d.append(Vec3d(v.co.x*scale.x, v.co.y*scale.y, v.co.z*scale.z))
			convexDesc.vertices = v_vec3d
			self.cdShape = spbapi.GetFWSdk().GetPHSdk().CreateShape(CDConvexMesh.GetIfInfoStatic(), convexDesc)

		if self.cdShape is None:
			print("Error : Error on Creating Shape <!!>")
			return

		# Set Shape to Solid
		print("Set Shape.")
		self.solidHnd.phSolid.AddShape(self.cdShape)

		# Set Name
		self.cdShape.SetName("cd_"+self.object.name)

		# 依存ハンドラのセット
		self.dependent_handlers = [self.solidHnd]

		# 構築完了フラグを更新
		self.build_finished = ( (not self.cdShape is None) and (not self.solidHnd is None) )

	def destroy(self):
		if (not self.cdShape is None) and (not self.solidHnd is None) and (not self.solidHnd.phSolid is None):
			self.solidHnd.phSolid.RemoveShape(self.cdShape)
			self.cdShape = None
		self.build_finished = False

	@classmethod
	def is_target(cls, object):
		'''オブジェクトがこのハンドラの対象となるべきものであるかを返す'''
		solid_type  = (object.spr_object_type=="Solid")
		has_shape   = (not object.spr_shape_type=="None")
		return solid_type and has_shape


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class ShapePoseSync(Synchronizer):
	def __init__(self, handler):
		Synchronizer.__init__(self, handler)
		self.bpy = to_bpy([None, None])
		self.spr = to_spr([None, None])

	def is_bpy_changed(self):
		last_relative = self.bpy
		mat_relative  = self.get_bpy_value_mod(self.bpy)
		if mat_relative is None or last_relative is None:
			return True

		mat_diff     = mat_relative - last_relative
		mat_diff_abs = 0
		for i in range(0,4):
			for j in range(0,4):
				mat_diff_abs += abs(mat_diff[i][j])
		return mat_diff_abs > 1e-5

	def is_spr_changed(self):
		# あとで実装
		return False

	def get_bpy_value(self, last):
		# -- 親子それぞれのWorld座標から，相対変換を求める --
		mat_child    = self.handler.object.matrix_world
		mat_parent   = self.handler.solidHnd.object.matrix_world
		mat_relative = mat_parent.inverted() * mat_child
		return mat_relative

	def get_spr_value(self, last):
		# あとで実装
		return None

	def set_bpy_value(self, value):
		# あとで実装
		pass

	def set_spr_value(self, value):
		solidHnd = self.handler.solidHnd
		if solidHnd is None or solidHnd.phSolid is None:
			return
		
		# Shape番号を探してきてSetShapePoseする
		mat_relative = value
		for i in range(0, solidHnd.phSolid.NShape()):
			if solidHnd.phSolid.GetShape(i)==self.handler.cdShape:
				solidHnd.phSolid.SetShapePose(i, mat_relative)

		# Solidハンドラに形状に変化があったことを伝える　→　慣性テンソルの再計算が行われる．
		solidHnd.shape_changed = True
		

class ShapeDimensionSync(Synchronizer):
	def get_bpy_value(self, last):
		return self.handler.object.dimensions
	def set_spr_value(self, value):
		changed = False
		if self.handler.build_info["shapetype"] == "Sphere":
			radius = (value.x + value.y + value.z) / 3 / 2
			self.handler.cdShape.SetRadius(radius)
			changed = True
		if self.handler.build_info["shapetype"] == "Box":
			boxsize = Vec3d(value.x, value.y, value.z)
			self.handler.cdShape.SetBoxSize(boxsize)
			changed = True

		# Solidハンドラに形状に変化があったことを伝える　→　慣性テンソルの再計算が行われる．
		if changed:
			solidHnd = self.handler.solidHnd
			if solidHnd is None or solidHnd.phSolid is None:
				return
			solidHnd.shape_changed = True


class ShapeElasticitySync(Synchronizer):
	def get_bpy_value(self, last):
		mat = self.handler.object.material_slots.id_data.active_material
		return mat.physics.elasticity if (not mat is None) else self.handler.cdShape.GetElasticity()
	def get_spr_value(self, last):
		return self.handler.cdShape.GetElasticity()
	def set_bpy_value(self, value):
		mat = self.handler.object.material_slots.id_data.active_material
		if not mat is None:
			mat.physics.elasticity = value
	def set_spr_value(self, value):
		self.handler.cdShape.SetElasticity(value)


class ShapeFrictionSync(Synchronizer):
	def get_bpy_value(self, last):
		mat = self.handler.object.material_slots.id_data.active_material
		return mat.physics.friction if (not mat is None) else self.handler.cdShape.GetDynamicFriction()
	def get_spr_value(self, last):
		return self.handler.cdShape.GetDynamicFriction()
	def set_bpy_value(self, value):
		mat = self.handler.object.material_slots.id_data.active_material
		if not mat is None:
			mat.physics.friction = value
	def set_spr_value(self, value):
		self.handler.cdShape.SetDynamicFriction(value)


class ShapeStaticFrictionSync(Synchronizer):
	def get_bpy_value(self, last):
		mat = self.handler.object.material_slots.id_data.active_material
		return mat.spr_friction_static if (not mat is None) else self.handler.cdShape.GetStaticFriction()
	def get_spr_value(self, last):
		return self.handler.cdShape.GetStaticFriction()
	def set_bpy_value(self, value):
		mat = self.handler.object.material_slots.id_data.active_material
		if not mat is None:
			mat.spr_friction_static = value
	def set_spr_value(self, value):
		self.handler.cdShape.SetStaticFriction(value)
