# -*- coding: utf-8 -*-

import bpy
import spb

from   math             import sin, cos, tan, pi
from   mathutils        import Vector,Quaternion
from   spb.handler      import Handler
from   spb.spbapi       import spbapi

from   Spr              import *
from   spb.synchronizer import *
from   spb.utils        import *


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Handler
#

class VisualizerHandler(Handler):
	'''個々のビジュアライザに対応するハンドラ．'''

	def __init__(self, group, name):
		Handler.__init__(self)

		self.group = group
		self.name  = name

		self.visualizer = None

		# -- Synchronizers
		self.visualizer_sync = VisualizerSync(self)
		self.show_sync       = VisualizerShowSync(self)

	def setting(self):
		return self.group.spr_visualizers[self.name]

	# --- --- --- --- ---
	# 基本クラスのオーバーライド

	def update_build_info(self):
		'''Springheadオブジェクトの構築に必要な情報を更新．これが変化すると再構築される'''
		# ...とはいうもののVisualizerの場合構築されるのはbpyオブジェクトとかだったりして．
		self.build_info["type"] = self.setting().type

	def build(self):
		'''対応するSpringhead側オブジェクトを構築する．依存ハンドラの取得も行う．'''
		# ...と言っているけどやはりVisualizerの場合構築されるのはbpyオブジェクトとかだったりして．
		type = self.build_info["type"]

		# 旧オブジェクトがあれば消しておく
		self.destroy()

		# ビジュアライザを構築
		if type=="FOV":
			self.visualizer = FOVVisualizer(self)
		if type=="Attention":
			self.visualizer = AttentionVisualizer(self)

		# 構築完了フラグを更新
		self.build_finished = (not self.visualizer is None)

	def destroy(self):
		'''Springheadオブジェクトを削除する．存在しないものものを消さないように気をつけること'''
		# しつこいようですがVisualizerの場合はSpringheadオブジェクトではありません．
		if not self.visualizer is None:
			self.visualizer.remove()
			self.visualizer = None
		self.build_finished = False

	def must_removed(self):
		'''このハンドラはもはや消滅すべき状態なのではないだろうか'''
		for vis in self.group.spr_visualizers:
			if vis.name == self.name:
				return False
		return True

	def record(self, frame):
		'''キーフレームへの記録を行う．'''
		self.visualizer.record(frame)



# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Synchronizer
#

class VisualizerSync(Synchronizer):
	def get_spr_value(self, last):
		return self.handler.visualizer.get()
	def set_bpy_value(self, value):
		self.handler.visualizer.set(value)


class VisualizerShowSync(Synchronizer):
	def get_spr_value(self, last):
		return self.handler.setting().show
	def set_bpy_value(self, value):
		if value:
			self.handler.visualizer.show()
		else:
			self.handler.visualizer.hide()


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---
#
# Visualizer
#

class FOVVisualizer():
	def __init__(self, handler):
		self.object = None
		self.handler = handler
		self.name    = self.handler.setting().name
		self.get()
		self.prepare_object()

	def get(self):
		self.param_name_horiz = self.handler.setting().fov_param_name_horiz
		self.param_name_vert  = self.handler.setting().fov_param_name_vert
		horiz  = get_cr_param(self.handler.group, self.param_name_horiz, default=0)
		vert   = get_cr_param(self.handler.group, self.param_name_vert,  default=0)
		target = self.handler.setting().target
		return [horiz, vert, self.param_name_horiz, self.param_name_vert, target]

	def set(self, value):
		horiz, vert, horizname, vertname, target = value
		x = tan(rad(horiz/2))
		y = tan(rad(vert/2))
		self.object.scale = (x, y, 1)
		if target in bpy.data.objects:
			self.object.parent = bpy.data.objects[target]

	def remove(self):
		if (not self.object==None) and (not self.object.name==None) and (self.object.name in bpy.data.objects):
			object = bpy.data.objects[self.object.name]
			bpy.context.scene.objects.unlink(object)
			bpy.data.objects.remove(object)

	def record(self, frame):
		'''キーフレームへの記録を行う．'''
		spb.handlers.keys.insert(self.object, frame, location=True, rotation=True, scale=True)

	# ----- ----- ----- ----- ----- ----- ----- ----- ----- -----

	def show(self):
		if not self.object==None:
			self.object.hide = False

	def hide(self):
		if not self.object==None:
			self.object.hide = True

	def prepare_object(self):
		active_object = bpy.context.scene.objects.active
		obj_name = "FOVConeVisualizerObj." + self.name
		mat_name = "FOVConeVisualizerMat." + self.name
		try:
			self.object = BpyObject(bpy.data.objects[obj_name])
		except:
			# -- Create New Cone --
			depth        = 10.0
			vertices_num = 32
			radius_alt   = 10.0
			radius_az    = 10.0

			#2D array {[0,0,0],[0,0,0],...}
			vertices = [[0 for i in range(0, 3, 1)] for j in range(0, vertices_num + 2, 1)]

			vertices[0] = (0,0,0)
			i = 1
			while i <= vertices_num:
				vertices[i][0] =  radius_az  * cos(i * pi/(vertices_num/2))
				vertices[i][1] =  radius_alt * sin(i * pi/(vertices_num/2))
				vertices[i][2] = -depth
				i += 1
				
			vertices[vertices_num + 1] = (0, 0, -depth)
				
			i = 0
			faces = [[0 for i in range(0, 3, 1)] for j in range(0, vertices_num*2, 1)]
			while i < (vertices_num - 1):
				faces[i]                = (0,   i+1, i+2)
				faces[vertices_num + i] = (i+2, i+1, vertices_num+1)
				i += 1
				
			faces[vertices_num   - 1] = (0, vertices_num, 1)
			faces[vertices_num*2 - 1] = (1, vertices_num, vertices_num+1)
				
			self.object = BpyObject(create_mesh(obj_name, (0,0,0), vertices, [], faces))
				
			self.object.rotation_euler = (rad(90.0), 0.0, rad(180))
				
			#描画関連
			bpy.ops.object.select_all(action='DESELECT')
			self.object.select = True
			bpy.ops.object.shade_smooth()
			
			self.object.show_x_ray  = True
			self.object.hide_select = True
				
			#Set Material
			mat = prepare_material(mat_name, (0,1,0), (0.5,0.5,0), 0.2)
			set_material(self.object, mat)

		self.object.select = False
		bpy.context.scene.objects.active = active_object


# --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- --- ---

class AttentionVisualizer():
	def __init__(self, handler):
		self.handler = handler
		self.name    = self.handler.setting().name
		self.items   = NoneCheckDict()

	def update(self, obj_name, vAtt, tAtt, tdAtt, totalAtt):
		if obj_name in bpy.data.objects:
			item   = self.items[obj_name]
			if item is None:
				item = AttentionVisualizerItem(self.handler.group.name+"_"+self.name+"_"+obj_name)
				item.attach_to(bpy.data.objects[obj_name])
				self.items[obj_name] = item
			item.update(vAtt, tAtt, tdAtt, totalAtt)

	def remove(self):
		for item in self.items.values():
			item.remove()

	def show(self):
		for item in self.items.values():
			item.show()

	def hide(self):
		for item in self.items.values():
			item.hide()

	def get(self):
		return None

	def set(self, value):
		pass

	def record(self, frame):
		'''キーフレームへの記録を行う．'''
		for item in self.items.values():
			item.record(frame)


# ----- ----- ----- ----- ----- ----- ----- ----- ----- -----

class AttentionVisualizerItem():
	def __init__(self, name):
		self.name   = name

		self.vObject     = None
		self.tObject     = None
		self.tdObject    = None
		self.totalObject = None
		self.objects = []

		self.target      = None

		self.vAtt     = 0.0
		self.tAtt     = 0.0
		self.tdAtt    = 0.0
		self.totalAtt = 0.0

		self.colors = {"Visual":(1,0,1), "Touch":(1,1,0), "TopDown":(0,1,1), "Total":(1,1,1)}

		self.prepare_objects()

	def attach_to(self, target):
		self.target = target
		for object in self.objects:
			object.parent = target

	def update(self, vAtt, tAtt, tdAtt, totalAtt):
		ps = Vector((1,1,1))
		if not vAtt==self.vAtt:
			s = 1 + vAtt
			# ps = self.target.scale
			self.vObject.scale = (s/ps.x, s/ps.y, s/ps.z)
			self.vAtt = vAtt

		if not tAtt==self.tAtt:
			s = 1 + tAtt
			# ps = self.target.scale
			self.tObject.scale = (s/ps.x, s/ps.y, s/ps.z)
			self.tAtt = tAtt

		if not tdAtt==self.tdAtt:
			s = 1 + tdAtt
			# ps = self.target.scale
			self.tdObject.scale = (s/ps.x, s/ps.y, s/ps.z)
			self.tdAtt = tdAtt

		if not totalAtt==self.totalAtt:
			s = 1 + totalAtt
			# ps = self.target.scale
			self.totalObject.scale = (s/ps.x, s/ps.y, s/ps.z)
			self.totalAtt = totalAtt

	# ----- ----- ----- ----- ----- ----- ----- ----- ----- -----

	def show(self):
		for object in self.objects:
			if not object==None:
				object.hide = False

	def hide(self):
		for object in self.objects:
			if not object==None:
				object.hide = True

	def remove(self):
		for object in self.objects:
			if (not object==None) and (not object.name==None) and (object.name in bpy.data.objects):
				o = bpy.data.objects[object.name]
				bpy.context.scene.objects.unlink(o)
				bpy.data.objects.remove(o)

	def record(self, frame):
		'''キーフレームへの記録を行う．'''
		for object in self.objects:
			spb.handlers.keys.insert(object, frame, location=True, rotation=True, scale=True)

	def prepare_objects(self):
		active_object = bpy.context.scene.objects.active

		self.objects = []

		self.vObject = self.prepare_object("Visual")
		self.objects.append(self.vObject)

		self.tObject = self.prepare_object("Touch")
		self.objects.append(self.tObject)

		self.tdObject = self.prepare_object("TopDown")
		self.objects.append(self.tdObject)

		self.totalObject = self.prepare_object("Total")
		self.objects.append(self.totalObject)

		for o in self.objects:
			o.select = False
		bpy.context.scene.objects.active = active_object

	def prepare_object(self, kind):
		obj_name = "AttentionVisualizerObj." + self.name + "." + kind
		mat_name = "AttentionVisualizerMat." + kind
		try:
			return BpyObject(bpy.data.objects[obj_name])
		except:
			# -- Create New Sphere --
			radius = 0.1
			bpy.ops.mesh.primitive_uv_sphere_add(size=0.1)
			obj = bpy.context.object
			obj.location = Vector((0,0,0))
			obj.name = obj_name

			#描画関連
			bpy.ops.object.shade_smooth()
			obj.show_x_ray  = True #Trueにするとbuffer shadowがでなくなる・・・
			obj.hide_select = True
				
			#Set Material
			mat = prepare_material(mat_name, self.colors[kind], (0.5,0.5,0), 0.4)
			set_material(obj, mat)
			return BpyObject(obj)
