# -*- coding: utf-8 -*-

from   spb.utils    import *


class Synchronizer():
	'''BlenderとSpringheadの同期を担う機能のベースクラス'''

	def __init__(self, handler):
		self.handler  = handler
		self.bpy      = None
		self.spr      = None

		# Sync中断モード
		# 0 : 中断しない
		# 1 : 1回だけsyncして中断
		# 2 : 中断
		self.hold_mode = 0

		# 他のSynchronizerが値を変更したい場合に使うmodifier（例：spring/damper ratio）
		self.modifiers = []

		# 持ち主であるhandlerのsynchronizerリストに登録
		self.handler.synchronizers.append(self)

	def storable(self):
		'''storableでないことを返すための関数 (StorableSynchronizerでオーバーライドされる)'''
		return False

	def get_bpy_value_mod(self, last):
		'''bpyの新しい値を取得し，modifierを適用して返す'''
		bpy = self.get_bpy_value(last)
		for bpy_mod, spr_mod in self.modifiers:
			bpy = bpy_mod(bpy)
		return bpy
			
	def get_spr_value_mod(self, last):
		'''sprの新しい値を取得し，modifierを適用して返す'''
		spr = self.get_spr_value(last)
		for bpy_mod, spr_mod in self.modifiers:
			spr = spr_mod(spr)
		return spr

	def bpy_sync_cond(self):
		'''bpy->sprのsyncを実行する条件：syncが有効（holdされてない）かつbpyに変更があるとき．'''
		hold = (self.hold_mode==2)
		if self.hold_mode==1:
			self.hold_mode = 2
		return (not hold) and (self.is_bpy_changed())

	def hold(self, v=True):
		'''Hold状態にする'''
		self.hold_mode = 2 if v else 0

	def apply(self):
		'''Hold中なら1回だけsyncする'''
		self.hold_mode = 1 if self.hold_mode==2 else self.hold_mode

	# -- 以下オーバーライド用

	def is_bpy_changed(self):
		'''bpyの新しい値を取得して，変更があったかを返す'''
		return( not self.bpy==self.get_bpy_value_mod(self.bpy) )
			
	def is_spr_changed(self):
		'''sprの新しい値を取得して，変更があったかを返す'''
		return( not self.spr==self.get_spr_value_mod(self.spr) )

	def get_bpy_value(self, last):
		'''bpyの新しい値を取得する'''
		return None
			
	def get_spr_value(self, last):
		'''sprの新しい値を取得する'''
		return None

	def set_bpy_value(self, value):
		'''sprに変更があった場合にbpyに反映する'''
		pass

	def set_spr_value(self, value):
		'''bpyに変更があった場合にsprに反映する'''
		pass



class StorableSynchronizer(Synchronizer):
	'''Store / Restoreに対応したシンクロナイザ．'''

	def __init__(self, handler):
		self.initialized  = NoneCheckDict()
		self.store_buffer = {}
		Synchronizer.__init__(self, handler)

	def storable(self):
		'''storableであることを返す'''
		return True

	def store(self, key, init=False):
		'''状態の一時保存を行う．'''
		# init==Trueの場合は，そのKeyへのstoreが初めて呼ばれた場合だけstoreする
		if init==False or (init==True and (self.initialized[key] is None)):
			self.store_buffer[key] = (self.bpy, self.spr)
			self.initialized[key] = True

	def restore(self, key):
		'''保存しておいた状態に戻す．'''
		if key in self.store_buffer:
			bpy, spr = self.store_buffer[key]
			self.set_bpy_value(bpy)
			self.set_spr_value(spr)


