# -*- coding: utf-8 -*-

import bpy
import threading
import traceback
import time

from   mathutils    import Vector,Quaternion, Matrix
from   Spr          import *
from   collections  import defaultdict



def BpyObject(obj):
	'''BpyObjectでラップして返す．ラッパ不使用モードではオブジェクトをそのまま帰す'''
	if bpy.context.scene.spr_use_object_wrapper==1:
		return BpyObjectClass(obj)
	else:
		obj.alive = True
		return obj


def BpyObjectRefresh():
	'''ラッパ使用時のみBpyObjectのキャッシュをクリアする関数'''
	if bpy.context.scene.spr_use_object_wrapper==1:
		BpyObjectClass.refresh()


class BpyObjectClass():
	'''bpy.data.objectのラッパ．定期的に名前から引き直すことでundo時のNull Referenceを防ぐ．少々重いらしい．'''
	obj_names = {}
	obj_cache = defaultdict(lambda: None)
	val_cache = defaultdict(lambda: defaultdict(lambda:None))

	@classmethod
	def refresh(cls):
		'''キャッシュをクリアする．定期的に呼ぶこと．'''
		cls.obj_cache.clear()

	def __init__(self, obj):
		BpyObjectClass.obj_names[self] = obj.name

	def __del__(self):
		BpyObjectClass.obj_names.pop(self)
		BpyObjectClass.obj_cache.pop(self)
		BpyObjectClass.val_cache.pop(self)

	def __getattr__(self, propname):
		if propname=="alive":
			return BpyObjectClass.obj_names[self] in bpy.data.objects
		else:
			o = BpyObjectClass.obj_cache[self]

			# -- キャッシュがなくなっている場合
			if o is None:
				# -- すでに消滅しているオブジェクトの場合，キャッシュしておいた値を返す．
				name = BpyObjectClass.obj_names[self]
				if not name in bpy.data.objects:
					return BpyObjectClass.val_cache[self][propname]

				# -- 名前から取得する．
				o = bpy.data.objects[name]
				BpyObjectClass.obj_cache[self] = o

			val = getattr(o, propname)
			BpyObjectClass.val_cache[self][propname] = val
			return val

	def __setattr__(self, propname, value):
		o = BpyObjectClass.obj_cache[self]

		# -- キャッシュがなくなっている場合
		if o is None:
			# -- すでに消滅しているオブジェクトの場合は何もしない．
			name = BpyObjectClass.obj_names[self]
			if not name in bpy.data.objects:
				return None

			# -- 名前から取得する．
			o = bpy.data.objects[name]
			BpyObjectClass.obj_cache[self] = o

		return setattr(o, propname, value)


class BpyGroup():
	'''bpy.data.groupのラッパ．使用時に名前から引くことでundo時のNull Referenceを防ぐ．'''
	grpnames = {}
	lastvals = defaultdict(lambda: defaultdict(lambda:None))

	def __init__(self, grp):
		BpyGroup.grpnames[self] = grp.name

	def __del__(self):
		BpyGroup.grpnames.pop(self)
		BpyGroup.lastvals.pop(self)

	def __getattr__(self, propname):
		if propname=="alive":
			return BpyGroup.grpnames[self] in bpy.data.groups
		elif BpyGroup.grpnames[self] in bpy.data.groups:
			# -- bpy.data.groupsからグループを取得し，getattrして返す
			g = bpy.data.groups[BpyGroup.grpnames[self]]
			val = getattr(g, propname)
			BpyGroup.lastvals[BpyGroup.grpnames[self]][propname] = val
			return val
		else:
			# -- 消滅している場合，一回前の値を返す．
			return BpyGroup.lastvals[BpyGroup.grpnames[self]][propname]

	def __setattr__(self, propname, value):
		if BpyGroup.grpnames[self] in bpy.data.groups:
			# -- bpy.data.groupsからグループを取得し，setattrする
			g = bpy.data.groups[BpyGroup.grpnames[self]]
			return setattr(g, propname, value)
		else:
			return None


class NoneCheckDict(dict):
	'''要素が存在しない場合はNoneを返すdict．
	defaultdictと違い，存在しない要素にアクセスしても要素として追加はしない．'''

	def __getitem__(self, key):
		if key in self.keys():
			return dict.__getitem__(self, key)
		else:
			return None


def copy(v):
	'''どんなベクトル等であってもコピーする関数'''
	if type(v) is Vec2d:
		return Vec2d(v.x, v.y)

	if type(v) is Vec2f:
		return Vec2f(v.x, v.y)

	if type(v) is Vec3d:
		return Vec3d(v.x, v.y, v.z)

	if type(v) is Vec3f:
		return Vec3f(v.x, v.y, v.z)

	if type(v) is Quaterniond:
		return Quaterniond(v.w, v.x, v.y, v.z)

	if type(v) is Quaternionf:
		return Quaternionf(v.w, v.x, v.y, v.z)

	if type(v) is Posed:
		return Posed(copy(v.getPos()), copy(v.getOri()))

	if type(v) is Posef:
		return Posed(copy(v.getPos()), copy(v.getOri()))

	if type(v) is Vector:
		return v.copy()

	if type(v) is Quaternion:
		return v.copy()

	if type(v) is Matrix:
		return v.copy()

	if type(v) is list:
		r = []
		for i in v:
			r.append(copy(i))
		return r

	return v


def to_f(v):
	'''XXXd を XXXf にする'''
	if type(v) is Vec2d:
		return Vec2f(v.x, v.y)

	if type(v) is Vec3d:
		return Vec3f(v.x, v.y, v.z)

	if type(v) is Quaterniond:
		return Quaternionf(v.w, v.x, v.y, v.z)

	if type(v) is Posed:
		return Posef(to_f(v.getPos()), to_f(v.getOri()))

	if type(v) is list:
		r = []
		for i in v:
			r.append(to_f(i))
		return r

	return v


def to_d(v):
	'''XXXf を XXXd にする'''
	if type(v) is Vec2f:
		return Vec2d(v.x, v.y)

	if type(v) is Vec3f:
		return Vec3d(v.x, v.y, v.z)

	if type(v) is Quaternionf:
		return Quaterniond(v.w, v.x, v.y, v.z)

	if type(v) is Posef:
		return Posed(to_d(v.getPos()), to_d(v.getOri()))

	if type(v) is list:
		r = []
		for i in v:
			r.append(to_d(i))
		return r

	return v


def to_spr(v):
	'''bpyのベクトルやクォータニオンをsprに変換する関数'''
	if type(v) is Vector:
		return Vec3d(v.x, v.y, v.z)

	if type(v) is Quaternion:
		return Quaterniond(v.w, v.x, v.y, v.z)

	if type(v) is Matrix:
		p = v.to_translation()
		q = v.to_quaternion()
		pose = Posed(Vec3d(p.x,p.y,p.z), Quaterniond(q.w,q.x,q.y,q.z))
		return pose

	if type(v) is list:
		r = []
		for i in v:
			r.append(to_spr(i))
		return r

	return v


def to_bpy(v):
	'''sprのベクトルやクォータニオンをbpyに変換する関数'''
	if type(v) is Vec3d or type(v) is Vec3f:
		return Vector((v.x, v.y, v.z))

	if type(v) is Quaterniond or type(v) is Quaternionf:
		return Quaternion((v.w, v.x, v.y, v.z))

	if type(v) is Posed or type(v) is Posef:
		a = Affined()
		m = Matrix()
		v.ToAffine(a)
		for i in range(0,4):
			m[i][0] = a.row(i).w
			m[i][1] = a.row(i).x
			m[i][2] = a.row(i).y
			m[i][3] = a.row(i).z

	if type(v) is list:
		r = []
		for i in v:
			r.append(to_bpy(i))
		return r

	return v


def scale_matrix(s):
	'''ベクトル表現のscaleをAffine行列に変換する'''
	return Matrix(((s.x,0,0,0), (0,s.y,0,0), (0,0,s.z,0), (0,0,0,1)))


def rad(deg):
	'''度をラジアンへ変換'''
	return deg / 180.0 * 3.1415926


def deg(rad):
	'''ラジアンを度へ変換'''
	return rad * 180.0 / 3.1415926


def get_cr_param(group, name, default=None):
	'''クリーチャのパラメータをコレクションから取得'''
	for param in group.spr_creature_parameters:
		if param.name==name:
			return param.value
	print("Parameter Not Found : ", group.name, " : ", name)
	return default


def make_material(name, diffuse, specular, alpha):
	'''マテリアルを作成する'''
	mat = bpy.data.materials.new(name)
	mat.diffuse_color = diffuse
	mat.diffuse_shader = 'LAMBERT' 
	mat.diffuse_intensity = 1.0 
	mat.specular_color = specular
	mat.specular_shader = 'COOKTORR'
	mat.specular_intensity = 0.5
	mat.alpha = alpha
	mat.ambient = 1
	return mat


def prepare_material(mat_name, diffuse, specular, alpha):
	'''マテリアルを作成または取得（既存の場合）する'''
	try:
		mat = bpy.data.materials[mat_name]
	except:
		mat = make_material(mat_name, diffuse, specular, alpha)
		mat.use_transparency		= 1
		mat.use_shadeless			= True
		mat.use_cast_buffer_shadows = False

	return mat


def set_material(ob, mat):
	'''マテリアルをセットする'''
	me = ob.data
	me.materials.append(mat)


def create_mesh(name, origin, verts, edges, faces):
	'''メッシュを作成する'''
	# Create mesh and object
	me = bpy.data.meshes.new(name+'Mesh')
	ob = bpy.data.objects.new(name, me)
	ob.location = origin
	ob.show_name = True
	# Link object to scene
	scn = bpy.context.scene
	scn.objects.link(ob)
	scn.objects.active = ob
	scn.update()

	# Create mesh from given verts, edges, faces. 
	# Either edges or faces should be [], or you ask for problems
	me.from_pydata(verts, edges, faces)

	# Update mesh with new data
	me.update(calc_edges=True)

	return ob
	

def show_icons(layout):
	'''layoutにアイコン一覧を表示するテスト用関数'''
	icons = ['NONE', 'QUESTION', 'ERROR', 'CANCEL', 'TRIA_RIGHT', 'TRIA_DOWN', 'TRIA_LEFT', 'TRIA_UP', 'ARROW_LEFTRIGHT', 'PLUS', 'DISCLOSURE_TRI_DOWN', 'DISCLOSURE_TRI_RIGHT', 'RADIOBUT_OFF', 'RADIOBUT_ON', 'MENU_PANEL', 'BLENDER', 'DOT', 'X', 'GO_LEFT', 'PLUG', 'UI', 'NODE', 'NODE_SEL', 'FULLSCREEN', 'SPLITSCREEN', 'RIGHTARROW_THIN', 'BORDERMOVE', 'VIEWZOOM', 'ZOOMIN', 'ZOOMOUT', 'PANEL_CLOSE', 'COPY_ID', 'EYEDROPPER', 'LINK_AREA', 'AUTO', 'CHECKBOX_DEHLT', 'CHECKBOX_HLT', 'UNLOCKED', 'LOCKED', 'UNPINNED', 'PINNED', 'SCREEN_BACK', 'RIGHTARROW', 'DOWNARROW_HLT', 'DOTSUP', 'DOTSDOWN', 'LINK', 'INLINK', 'PLUGIN', 'HELP', 'GHOST_ENABLED', 'COLOR', 'LINKED', 'UNLINKED', 'HAND', 'ZOOM_ALL', 'ZOOM_SELECTED', 'ZOOM_PREVIOUS', 'ZOOM_IN', 'ZOOM_OUT', 'RENDER_REGION', 'BORDER_RECT', 'BORDER_LASSO', 'FREEZE', 'STYLUS_PRESSURE', 'GHOST_DISABLED', 'NEW', 'FILE_TICK', 'QUIT', 'URL', 'RECOVER_LAST', 'FULLSCREEN_ENTER', 'FULLSCREEN_EXIT', 'BLANK1', 'LAMP', 'MATERIAL', 'TEXTURE', 'ANIM', 'WORLD', 'SCENE', 'EDIT', 'GAME', 'RADIO', 'SCRIPT', 'PARTICLES', 'PHYSICS', 'SPEAKER', 'TEXTURE_SHADED', 'VIEW3D', 'IPO', 'OOPS', 'BUTS', 'FILESEL', 'IMAGE_COL', 'INFO', 'SEQUENCE', 'TEXT', 'IMASEL', 'SOUND', 'ACTION', 'NLA', 'SCRIPTWIN', 'TIME', 'NODETREE', 'LOGIC', 'CONSOLE', 'PREFERENCES', 'CLIP', 'ASSET_MANAGER', 'OBJECT_DATAMODE', 'EDITMODE_HLT', 'FACESEL_HLT', 'VPAINT_HLT', 'TPAINT_HLT', 'WPAINT_HLT', 'SCULPTMODE_HLT', 'POSE_HLT', 'PARTICLEMODE', 'LIGHTPAINT', 'SCENE_DATA', 'RENDERLAYERS', 'WORLD_DATA', 'OBJECT_DATA', 'MESH_DATA', 'CURVE_DATA', 'META_DATA', 'LATTICE_DATA', 'LAMP_DATA', 'MATERIAL_DATA', 'TEXTURE_DATA', 'ANIM_DATA', 'CAMERA_DATA', 'PARTICLE_DATA', 'LIBRARY_DATA_DIRECT', 'GROUP', 'ARMATURE_DATA', 'POSE_DATA', 'BONE_DATA', 'CONSTRAINT', 'SHAPEKEY_DATA', 'CONSTRAINT_BONE', 'PACKAGE', 'UGLYPACKAGE', 'BRUSH_DATA', 'IMAGE_DATA', 'FILE', 'FCURVE', 'FONT_DATA', 'RENDER_RESULT', 'SURFACE_DATA', 'EMPTY_DATA', 'SETTINGS', 'RENDER_ANIMATION', 'RENDER_STILL', 'BOIDS', 'STRANDS', 'LIBRARY_DATA_INDIRECT', 'GREASEPENCIL', 'GROUP_BONE', 'GROUP_VERTEX', 'GROUP_VCOL', 'GROUP_UVS', 'RNA', 'RNA_ADD', 'OUTLINER_OB_EMPTY', 'OUTLINER_OB_MESH', 'OUTLINER_OB_CURVE', 'OUTLINER_OB_LATTICE', 'OUTLINER_OB_META', 'OUTLINER_OB_LAMP', 'OUTLINER_OB_CAMERA', 'OUTLINER_OB_ARMATURE', 'OUTLINER_OB_FONT', 'OUTLINER_OB_SURFACE', 'OUTLINER_OB_SPEAKER', 'RESTRICT_VIEW_OFF', 'RESTRICT_VIEW_ON', 'RESTRICT_SELECT_OFF', 'RESTRICT_SELECT_ON', 'RESTRICT_RENDER_OFF', 'RESTRICT_RENDER_ON', 'OUTLINER_DATA_EMPTY', 'OUTLINER_DATA_MESH', 'OUTLINER_DATA_CURVE', 'OUTLINER_DATA_LATTICE', 'OUTLINER_DATA_META', 'OUTLINER_DATA_LAMP', 'OUTLINER_DATA_CAMERA', 'OUTLINER_DATA_ARMATURE', 'OUTLINER_DATA_FONT', 'OUTLINER_DATA_SURFACE', 'OUTLINER_DATA_SPEAKER', 'OUTLINER_DATA_POSE', 'MESH_PLANE', 'MESH_CUBE', 'MESH_CIRCLE', 'MESH_UVSPHERE', 'MESH_ICOSPHERE', 'MESH_GRID', 'MESH_MONKEY', 'MESH_CYLINDER', 'MESH_TORUS', 'MESH_CONE', 'LAMP_POINT', 'LAMP_SUN', 'LAMP_SPOT', 'LAMP_HEMI', 'LAMP_AREA', 'META_EMPTY', 'META_PLANE', 'META_CUBE', 'META_BALL', 'META_ELLIPSOID', 'META_CAPSULE', 'SURFACE_NCURVE', 'SURFACE_NCIRCLE', 'SURFACE_NSURFACE', 'SURFACE_NCYLINDER', 'SURFACE_NSPHERE', 'SURFACE_NTORUS', 'CURVE_BEZCURVE', 'CURVE_BEZCIRCLE', 'CURVE_NCURVE', 'CURVE_NCIRCLE', 'CURVE_PATH', 'FORCE_FORCE', 'FORCE_WIND', 'FORCE_VORTEX', 'FORCE_MAGNETIC', 'FORCE_HARMONIC', 'FORCE_CHARGE', 'FORCE_LENNARDJONES', 'FORCE_TEXTURE', 'FORCE_CURVE', 'FORCE_BOID', 'FORCE_TURBULENCE', 'FORCE_DRAG', 'MODIFIER', 'MOD_WAVE', 'MOD_BUILD', 'MOD_DECIM', 'MOD_MIRROR', 'MOD_SOFT', 'MOD_SUBSURF', 'HOOK', 'MOD_PHYSICS', 'MOD_PARTICLES', 'MOD_BOOLEAN', 'MOD_EDGESPLIT', 'MOD_ARRAY', 'MOD_UVPROJECT', 'MOD_DISPLACE', 'MOD_CURVE', 'MOD_LATTICE', 'CONSTRAINT_DATA', 'MOD_ARMATURE', 'MOD_SHRINKWRAP', 'MOD_CAST', 'MOD_MESHDEFORM', 'MOD_BEVEL', 'MOD_SMOOTH', 'MOD_SIMPLEDEFORM', 'MOD_MASK', 'MOD_CLOTH', 'MOD_EXPLODE', 'MOD_FLUIDSIM', 'MOD_MULTIRES', 'MOD_SMOKE', 'MOD_SOLIDIFY', 'MOD_SCREW', 'MOD_VERTEX_WEIGHT', 'MOD_DYNAMICPAINT', 'MOD_REMESH', 'REC', 'PLAY', 'FF', 'REW', 'PAUSE', 'PREV_KEYFRAME', 'NEXT_KEYFRAME', 'PLAY_AUDIO', 'PLAY_REVERSE', 'PREVIEW_RANGE', 'PMARKER_ACT', 'PMARKER_SEL', 'PMARKER', 'MARKER_HLT', 'MARKER', 'SPACE2', 'SPACE3', 'KEYINGSET', 'KEY_DEHLT', 'KEY_HLT', 'MUTE_IPO_OFF', 'MUTE_IPO_ON', 'VISIBLE_IPO_OFF', 'VISIBLE_IPO_ON', 'DRIVER', 'SOLO_OFF', 'SOLO_ON', 'FRAME_PREV', 'FRAME_NEXT', 'VERTEXSEL', 'EDGESEL', 'FACESEL', 'ROTATE', 'CURSOR', 'ROTATECOLLECTION', 'ROTATECENTER', 'ROTACTIVE', 'ALIGN', 'SMOOTHCURVE', 'SPHERECURVE', 'ROOTCURVE', 'SHARPCURVE', 'LINCURVE', 'NOCURVE', 'RNDCURVE', 'PROP_OFF', 'PROP_ON', 'PROP_CON', 'PARTICLE_POINT', 'PARTICLE_TIP', 'PARTICLE_PATH', 'MAN_TRANS', 'MAN_ROT', 'MAN_SCALE', 'MANIPUL', 'SNAP_OFF', 'SNAP_ON', 'SNAP_NORMAL', 'SNAP_INCREMENT', 'SNAP_VERTEX', 'SNAP_EDGE', 'SNAP_FACE', 'SNAP_VOLUME', 'STICKY_UVS_LOC', 'STICKY_UVS_DISABLE', 'STICKY_UVS_VERT', 'CLIPUV_DEHLT', 'CLIPUV_HLT', 'SNAP_PEEL_OBJECT', 'GRID', 'PASTEDOWN', 'COPYDOWN', 'PASTEFLIPUP', 'PASTEFLIPDOWN', 'SNAP_SURFACE', 'RETOPO', 'UV_VERTEXSEL', 'UV_EDGESEL', 'UV_FACESEL', 'UV_ISLANDSEL', 'UV_SYNC_SELECT', 'BBOX', 'WIRE', 'SOLID', 'SMOOTH', 'POTATO', 'ORTHO', 'LOCKVIEW_OFF', 'LOCKVIEW_ON', 'AXIS_SIDE', 'AXIS_FRONT', 'AXIS_TOP', 'NDOF_DOM', 'NDOF_TURN', 'NDOF_FLY', 'NDOF_TRANS', 'LAYER_USED', 'LAYER_ACTIVE', 'SORTALPHA', 'SORTBYEXT', 'SORTTIME', 'SORTSIZE', 'LONGDISPLAY', 'SHORTDISPLAY', 'GHOST', 'IMGDISPLAY', 'BOOKMARKS', 'FONTPREVIEW', 'FILTER', 'NEWFOLDER', 'FILE_PARENT', 'FILE_REFRESH', 'FILE_FOLDER', 'FILE_BLANK', 'FILE_BLEND', 'FILE_IMAGE', 'FILE_MOVIE', 'FILE_SCRIPT', 'FILE_SOUND', 'FILE_FONT', 'BACK', 'FORWARD', 'DISK_DRIVE', 'MATPLANE', 'MATSPHERE', 'MATCUBE', 'MONKEY', 'HAIR', 'ALIASED', 'ANTIALIASED', 'MAT_SPHERE_SKY', 'WORDWRAP_OFF', 'WORDWRAP_ON', 'SYNTAX_OFF', 'SYNTAX_ON', 'LINENUMBERS_OFF', 'LINENUMBERS_ON', 'SCRIPTPLUGINS', 'SEQ_SEQUENCER', 'SEQ_PREVIEW', 'SEQ_LUMA_WAVEFORM', 'SEQ_CHROMA_SCOPE', 'SEQ_HISTOGRAM', 'SEQ_SPLITVIEW', 'IMAGE_RGB', 'IMAGE_RGB_ALPHA', 'IMAGE_ALPHA', 'IMAGE_ZDEPTH', 'IMAGEFILE', 'BRUSH_ADD', 'BRUSH_BLOB', 'BRUSH_BLUR', 'BRUSH_CLAY', 'BRUSH_CLONE', 'BRUSH_CREASE', 'BRUSH_DARKEN', 'BRUSH_FILL', 'BRUSH_FLATTEN', 'BRUSH_GRAB', 'BRUSH_INFLATE', 'BRUSH_LAYER', 'BRUSH_LIGHTEN', 'BRUSH_MIX', 'BRUSH_MULTIPLY', 'BRUSH_NUDGE', 'BRUSH_PINCH', 'BRUSH_SCRAPE', 'BRUSH_SCULPT_DRAW', 'BRUSH_SMEAR', 'BRUSH_SMOOTH', 'BRUSH_SNAKE_HOOK', 'BRUSH_SOFTEN', 'BRUSH_SUBTRACT', 'BRUSH_TEXDRAW', 'BRUSH_THUMB', 'BRUSH_ROTATE', 'BRUSH_VERTEXDRAW', 'VIEW3D_VEC', 'EDIT_VEC', 'EDITMODE_DEHLT', 'EDITMODE_HLT', 'DISCLOSURE_TRI_RIGHT_VEC', 'DISCLOSURE_TRI_DOWN_VEC', 'MOVE_UP_VEC', 'MOVE_DOWN_VEC', 'X_VEC', 'SMALL_TRI_RIGHT_VEC']
	for icon_ in icons:
	 	layout.label(text=icon_, icon = icon_)


class Profilers:
	'''実行時間計測用．'''

	def __init__(self):
		self.curr_id   = 0
		self.profilers = defaultdict(lambda: Profiler())

	def start(self, id=0):
		'''ループ内の時間計測開始地点に置く．'''
		self.curr_id = id
		self.profilers[id].start(id)

	def lap(self, id=None):
		'''各計測ポイントに置く．start->lap間, およびlap->lap間の時間が測定される．'''
		id = id or self.curr_id
		self.profilers[id].lap()


class Profiler:
	def __init__(self):
		self.count     = 0
		self.id        = 0
		self.time_sum  = defaultdict(lambda: 0)
		self.starttime = time.time()

	def start(self, prof_id=0):
		self.id     = 0
		self.count += 1
		self.last   = time.time()

		now = time.time()
		if now - self.starttime > 5:
			print(" -- ", prof_id, " -- ")
			for i in range(0, len(self.time_sum)):
				print(i, " : ", ((self.time_sum[i]/self.count)*1000000).__int__()/1000.0, "[ms]")
				self.time_sum[i] = 0
			self.starttime = now
			self.count = 0

	def lap(self):
		now = time.time()
		self.time_sum[self.id] += (now - self.last)
		self.last  = now
		self.id   += 1

profiler = Profilers()
profiler.start()

